import re
import os
import datetime
import inspect
import itertools
import contextlib
import multiprocessing
import functools

import pandas as pd
import numpy as np
import pyranges as pr

import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome
import handygenome.tools as tools
import handygenome.logutils as logutils
import handygenome.cnv.rdnacopy as rdnacopy

import handygenome.genomedf.genomedf_utils as genomedf_utils
import handygenome.genomedf.genomedf_methods as genomedf_methods


##############################
# coordinate argument parser #
##############################

def coord_argparse_OLD(**kwargs):
    all_valid_keys = {
        'chrom': ['chrom', 'chroms', 'chromosome', 'chromosomes'],
        'start': ['start', 'starts', 'start0', 'start0s'],
        'end': ['end', 'ends', 'end0', 'end0s'],
    }

    # parse coordinates from raw arguments
    parsed_keys = dict()
    parsed_values = dict()
    for coordtype, valid_keys in all_valid_keys.items():
        selected_keys = [x for x in kwargs.keys() if x.lower() in valid_keys]
        if len(selected_keys) == 0:
            parsed_keys[coordtype] = None
            parsed_values[coordtype] = None
        elif len(selected_keys) == 1:
            parsed_keys[coordtype] = selected_keys[0]
            parsed_values[coordtype] = kwargs[selected_keys[0]]
            if parsed_values[coordtype] is not None:
                parsed_values[coordtype] = np.atleast_1d(parsed_values[coordtype])
        else:
            raise Exception(
                f'Among the argument keys, exactly one of '
                f'these (case-insensitive) must exist: {valid_keys}'
            )

    # sanitycheck
    if parsed_keys['chroms'] is None:
        raise Exception(
            f'Chromosomes must be specified by one of these '
            f'keywords (case-insensitive): {all_valid_keys["chroms"]}'
        )
    if (parsed_keys['start0s'] is None) and (parsed_keys['end0s'] is not None):
        raise Exception(f'Ends must not be given without starts.')

    # broadcast only not-None values
    not_none_coordtypes, not_none_values = zip(
        *(
            (coordtype, val) for (coordtype, val) in parsed_values.items() 
            if val is not None
        )
    )
    try:
        new_values = np.broadcast_arrays(*not_none_values)
    except ValueError as exc:
        if str(exc).startswith('shape mismatch'):
            myexc = Exception(
                f'"chroms", "start0s", "end0s" must be broadcastable to a single shape.'
            )
            raise myexc from exc
        else:
            raise

    for coordtype, new_val in zip(not_none_coordtypes, new_values):
        parsed_values[coordtype] = new_val

    # make chroms into 'object' type array (by default fixed length unicode)
    parsed_values['chroms'] = parsed_values['chroms'].astype(object)

    # prepare results
    coords = {
        'Chromosome': parsed_values['chroms'],
        'Start': parsed_values['start0s'],
        'End': parsed_values['end0s'],
    }
    coord_keys = set(parsed_keys.keys()).difference({None})
    other_kwargs = dict(
        (k, v) for (k, v) in kwargs.items()
        if k not in coord_keys
    )

    return coords, other_kwargs


COORDARG_MANDATORY_KEYS = ['refver', 'chroms', 'start0s', 'end0s']


def modify_coord_args(kwargs):
    """Modifies kwargs in-place. This is an intended behavior in order to modify "arguments" attribute of BoundArguments object."""
    if not set(COORDARG_MANDATORY_KEYS).issubset(kwargs.keys()):
        raise Exception(f'Mandatory keys: {COORDARG_MANDATORY_KEYS}')

    # sanitycheck
    if kwargs['chroms'] is None:
        if (
            (kwargs['start0s'] is not None) 
            or (kwargs['end0s'] is not None)
        ):
            raise Exception(f'Start and End must not be specified without Chromosomes')
    if (kwargs['start0s'] is None) and (kwargs['end0s'] is not None):
        raise Exception(f'Ends must not be given without starts.')

    # return without any change when chroms is not given
    if kwargs['chroms'] is None:
        return

    # transform chrom values
    kwargs['chroms'] = np.atleast_1d(kwargs['chroms']).astype(object)

    # fill in default start0s and end0s
    if kwargs['start0s'] is None:
        kwargs['start0s'] = 0
    if kwargs['end0s'] is None:
        chromdict = refgenome.get_chromdict(kwargs['refver'])
        kwargs['end0s'] = np.asarray([chromdict[x] for x in kwargs['chroms']])

    # broadcast
    try:
        kwargs['chroms'], kwargs['start0s'], kwargs['end0s'] = np.broadcast_arrays(
            kwargs['chroms'], kwargs['start0s'], kwargs['end0s'],
        )
    except ValueError as exc:
        if str(exc).startswith('shape mismatch'):
            myexc = Exception(
                f'"chroms", "start0s", "end0s" must be broadcastable to a single shape.'
            )
            raise myexc from exc
        else:
            raise


def deco_coordarg(func):
    sig = inspect.signature(func)
    if not set(COORDARG_MANDATORY_KEYS).issubset(sig.parameters.keys()):
        raise Exception(
            f'Parameters of the fuction {func.__name__} must include all of these: {COORDARG_MANDATORY_KEYS}'
        )

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ba = sig.bind(*args, **kwargs)
        ba.apply_defaults()
        modify_coord_args(ba.arguments)
        return func(*ba.args, **ba.kwargs)

    return wrapper


##############
# exceptions #
##############

class NotBinnedError(Exception):
    pass


#########################
# GenomeDataFrame class #
#########################

class GenomeDataFrame:
    ########
    # repr #
    ########

    def __repr__(self):
        return f'<{self.__class__.__name__} object>\n{repr(self.df)}'

    ################
    # constructors #
    ################

    def __init__(self, refver):
        self.refver = refgenome.standardize(refver)
        self._df = None
        self._df_lastaccess = datetime.datetime.fromtimestamp(0)
        self._gr = None
        self._gr_lastaccess = datetime.datetime.fromtimestamp(0)
        self._sorted = False
        #self._columns = None

    def __del__(self):
        del self._df
        del self._gr

    def spawn(self, frame, **kwargs):
        result = self.__class__(self.refver, **kwargs)
        result.assign_frame(frame, self.dtypes.to_dict())
        return result

    def spawn_from_data(self, **kwargs):
        kwargs['refver'] = self.refver
        return self.__class__.from_data(**kwargs)

    @classmethod
    def init_empty(cls, refver, annot_cols=tuple()):
        columns = genomedf_utils.COMMON_COLUMNS + list(annot_cols)
        df = pd.DataFrame([], columns=columns)
        return cls.from_frame(df, refver)

    def spawn_empty(self, annot_cols=tuple()):
        columns = genomedf_utils.COMMON_COLUMNS + list(annot_cols)
        df = pd.DataFrame([], columns=columns)
        return self.spawn(df)

    @classmethod
    def concat(cls, gdf_iterator):
        gdf_list = list(gdf_iterator)
        assert len(gdf_list) > 0

        refvers = set(x.refver for x in gdf_list)
        if len(refvers) != 1:
            raise Exception(f'Reference versions differ: {refvers}')
        refver = refvers.pop()

        new_df = pd.concat([x.df for x in gdf_list], axis=0)
        return cls.from_frame(new_df, refver)

    @classmethod
    def from_frame(cls, frame, refver, dtype=dict()):
        result = cls(refver)
        result.assign_frame(frame, dtype)
        return result

    @classmethod
    @deco_coordarg
    def from_data(cls, refver, chroms, start0s=None, end0s=None, **kwargs):
        # make source dict
        source_dict = dict()
        source_dict['Chromosome'] = chroms

        if start0s is None:
            source_dict['Start'] = np.atleast_1d(0)
        else:
            source_dict['Start'] = start0s

        if end0s is None:
            chromdict = refgenome.get_chromdict(refver)
            source_dict['End'] = [chromdict[x] for x in source_dict['Chromosome']]
        else:
            source_dict['End'] = end0s

        for key, val in kwargs.items():
            source_dict[key] = val

        df = pd.DataFrame(source_dict)

        return cls.from_frame(df, refver)

    @classmethod
    def from_records(cls, records, refver, annot_cols=tuple()):
        columns = genomedf_utils.COMMON_COLUMNS + list(annot_cols)
        df_source = dict()
        for key, val in itertools.zip_longest(columns, zip(*records), fillvalue=None):
            if (key is None) or (val is None):
                raise Exception(f'Length of annotation columns does not coincide between "records" and "annot_cols"')
            df_source[key] = val
        return cls.from_frame(pd.DataFrame(df_source), refver)

    @classmethod
    def all_regions(cls, refver, assembled_only=True):
        chromdict = refgenome.get_chromdict(refver)
        if assembled_only:
            chroms = refgenome.choose_assembled_chroms(chromdict.contigs)
        else:
            chroms = chromdict.contigs
        result = cls.from_data(refver=refver, chroms=chroms)
        return result

    @classmethod
    def from_margins(cls, refver, chrom_left, start0_left, chrom_right, end0_right):
        chromdict = refgenome.get_chromdict(refver)
        cumpos0_left = chromdict.get_cumpos0(chrom_left, start0_left)
        cumpos0_right = chromdict.get_cumpos0(chrom_right, end0_right)
        if cumpos0_left >= cumpos0_right:
            raise Exception(
                f'"left position" comes later than "right_position"; '
                f'chrom_left={chrom_left}, start0_left={start0_left}, '
                f'chrom_right={chrom_right}, end0_right={end0_right}'
            )

        #result = cls(refver)
        chroms = list()
        start0s = list()
        end0s = list()

        if chrom_left == chrom_right:
            chroms.append(chrom_left)
            start0s.append(start0_left)
            end0s.append(end0_right)
        else:
            chrom_left_idx = chromdict.contigs.index(chrom_left)
            chrom_right_idx = chromdict.contigs.index(chrom_right)

            chroms.append(chrom_left)
            start0s.append(start0_left)
            end0s.append(chromdict[chrom_left])

            for chrom in chromdict.contigs[(chrom_left_idx + 1):chrom_right_idx]:
                chroms.append(chrom)
                start0s.append(0)
                end0s.append(chromdict[chrom])

            chroms.append(chrom_right)
            start0s.append(0)
            end0s.append(end0_right)

        result = cls.from_data(
            refver=refver, chroms=chroms, start0s=start0s, end0s=end0s,
        )
        return result

    @staticmethod
    def from_data_argparse(kwargs, valid_keys):
        selected_keys = [x for x in kwargs.keys() if x.lower() in valid_keys]
        if len(selected_keys) == 0:
            return None
        elif len(selected_keys) == 1:
            return selected_keys[0]
        else:
            raise Exception(f'Among the argument keys, exactly one of these (case-insensitive) must exist: {valid_keys}')

    ###################
    # read/write file #
    ###################

    @classmethod
    def read_tsv(cls, filename, refver, annot_cols=None, dtype=dict()):
        frame = cls.make_frame_from_tsv(
            filename, annot_cols=annot_cols, dtype=dtype,
        )
        return cls.from_frame(frame, refver)

    from_tsv = read_tsv

    def write_tsv(self, filename):
        patstring = r'(bed|tsv)(\.gz)?$'
        assert re.search(patstring, filename), f'File name must match {patstring}'
        self.df.to_csv(filename, sep='\t', header=True, index=False)

    to_tsv = write_tsv

    @classmethod
    def make_frame_from_tsv(cls, filename, annot_cols=None, dtype=dict()):
        with tools.openfile(filename, 'rt') as infile:
            firstline = next(infile).strip()
        firstline_sp = firstline.split('\t')
        assert len(firstline_sp) >= 3

        has_header = not (firstline_sp[1].isdecimal() and firstline_sp[2].isdecimal())
        if has_header:
            assert (
                (firstline_sp[0] == 'Chromosome')
                and (firstline_sp[1] == 'Start')
                and (firstline_sp[2] == 'End')
            )

        if (not has_header) and (annot_cols is None):
            result = pr.read_bed(filename)
        else:
            if has_header:
                if annot_cols is None:
                    skiprows = None
                    header = 0 
                    names = None
                else:
                    skiprows = 1
                    header = None
                    names = genomedf_utils.COMMON_COLUMNS + list(annot_cols)
            else:
                skiprows = 0
                header = None
                names = genomedf_utils.COMMON_COLUMNS + list(annot_cols)

            dtype = (
                {'Chromosome': 'string', 'Start': int, 'End': int}
                | dtype
            )
            result = pd.read_csv(
                filename, 
                sep='\t', 
                skiprows=skiprows,
                header=header,
                names=names,
                dtype=dtype,
            )

        return result

    ###########################################################
    # assignment of dataframe/pyranges objects data populator #
    ###########################################################

    def assign_df(self, df, dtype=dict()):
        assert isinstance(df, pd.DataFrame)

        self.sanitycheck_df(df)
        self._df = self.postprocess_df(df, dtype)

        self.update_df_lastaccess()

    def assign_gr(self, gr):
        assert isinstance(gr, pr.PyRanges)
        
        self.sanitycheck_gr(gr)
        self._gr = self.postprocess_gr(gr)

        self.update_gr_lastaccess()

    def assign_frame(self, frame, dtype=dict()):
        if isinstance(frame, pd.DataFrame):
            return self.assign_df(frame, dtype)
        elif isinstance(frame, pr.PyRanges):
            return self.assign_gr(frame)
        else:
            raise Exception(
                f'Input "frame" object must be either "pandas DataFrame" or "PyRanges" object.'
            )

    @staticmethod
    def sanitycheck_common(chromosomes, ends, chromdict):
        if not np.isin(chromosomes, chromdict.keys()).all():
            raise Exception(f'Unknown chromosome in the input frame')

        maxends = chromosomes.apply(lambda x: chromdict[x])
        if not (ends <= maxends).all():
            raise Exception(f'Some of "End" values are greater than the chromosome length.')

    @classmethod
    def sanitycheck_df(cls, df):
        if not set(genomedf_utils.COMMON_COLUMNS).issubset(df.columns):
            raise Exception(f'Required columns: {genomedf_utils.COMMON_COLUMNS}')

        assert (df['End'] > df['Start']).all()  # this works with 0-row DataFrame

        #self.sanitycheck_common(df['Chromosome'], df['End'], self.chromdict)

    @classmethod
    def postprocess_df(cls, df, dtype=dict()):
        # all processes work for 0-row dataframe

        # dtype setting
        astype_arg = genomedf_utils.DEFAULT_DTYPES | dtype
        astype_arg = {
            key: val for key, val in astype_arg.items()
            if key in df.columns
        }
        df = df.astype(astype_arg)

        # column reordering
        new_cols = genomedf_utils.COMMON_COLUMNS + [
            x for x in df.columns if x not in genomedf_utils.COMMON_COLUMNS
        ]
            # df.columns.difference(genomedf_utils.COMMON_COLUMNS).to_list() -> does not preserve ordering
        df = df.loc[:, new_cols]

        # remove index
        df = df.reset_index(drop=True, inplace=False)

        return df

    def sanitycheck_gr(self, gr):
        pass
        #self.sanitycheck_common(gr.Chromosome, gr.End, self.chromdict)

    @staticmethod
    def postprocess_gr(gr):
        if not gr.empty:
            with open(os.devnull, 'w') as devnull:
                with contextlib.redirect_stdout(devnull):
                    if gr.dtypes['Start'] == np.int32:
                        gr.Start = gr.Start.to_numpy().astype('int64')
                    if gr.dtypes['End'] == np.int32:
                        gr.End = gr.End.to_numpy().astype('int64')

        return gr

    ################################
    # df, gr accessors and setters #
    ################################

    @property
    def df(self):
        if self._df is None:
            self.set_df_from_gr()
        else:
            if self._gr is not None:
                if self.check_gr_is_newer():
                    if not self.check_df_gr_same():
                        self.set_df_from_gr()

        self.update_df_lastaccess()
        return self._df

    @df.setter
    def df(self, value):
        raise Exception(f'Direct assignment of "df" attribute is not allowed.')

    @property
    def gr(self):
        if self._gr is None:
            self.set_gr_from_df()
        else:
            if self._df is not None:
                if not self.check_gr_is_newer():  # df is newer
                    if not self.check_df_gr_same():
                        self.set_gr_from_df()

        self.update_gr_lastaccess()
        return self._gr

    @gr.setter
    def gr(self, value):
        raise Exception(f'Direct assignment of "gr" attribute is not allowed.')

    #####################
    # helpers of df, gr #
    #####################

    @staticmethod
    def make_empty_df():
        result = pd.DataFrame([], columns=genomedf_utils.COMMON_COLUMNS)
        result = result.astype(genomedf_utils.DEFAULT_DTYPES)
        return result

    def set_gr_from_df(self):
        assert self._df is not None
        self._gr = pr.PyRanges(self._df, int64=True)

    def set_df_from_gr(self):
        assert self._gr is not None
        if self._gr.empty:
            if self._df is None:
                self._df = self.make_empty_df()
            else:
                self._df = self.df.iloc[[], :]
        else:
            self._df = self._gr.df.reset_index(drop=True)
                # "df" method of PyRanges object returns a newly created DataFrame
                # (gr.df is gr.df) is False

    def sync_df_gr(self):
        if not self.check_df_gr_same():
            if self.check_gr_is_newer():
                self.set_df_from_gr()
            else:
                self.set_gr_from_df()

    def update_df(self):
        if self.check_gr_is_newer():
            if not self.check_df_gr_same():
                self.set_df_from_gr()

    def update_gr(self):
        if not self.check_gr_is_newer():
            if not self.check_df_gr_same():
                self.set_gr_from_df()

    def update_df_lastaccess(self):
        self._df_lastaccess = datetime.datetime.now()

    def update_gr_lastaccess(self):
        self._gr_lastaccess = datetime.datetime.now()

    def check_gr_is_newer(self):
        return self._gr_lastaccess > self._df_lastaccess

    def check_df_gr_same(self):
        """This does not consider the order of rows because hash is calculated after sorting with genome coordinates."""
        assert self._df is not None
        assert self._gr is not None

        # when gr or df is empty
        gr_is_empty = self._gr.empty
        df_is_empty = (self._df.shape[0] == 0)
        if gr_is_empty or df_is_empty:
            return (gr_is_empty == df_is_empty)

        # both gr and df are not empty
        if set(self._gr.columns) != set(self._df.columns):
            return False
        else:
            df_for_hash = self._df.loc[:, self._gr.columns]  # make column orders the same
            return (
                genomedf_utils.hash_gr(self._gr) 
                == genomedf_utils.hash_df(df_for_hash)
            ).to_numpy().all()  
                # It is faster (~2.5 us vs ~13 us) to convert into numpy and then do .all()

    def get_newer_frame(self):
        if self.check_gr_is_newer():
            assert self._gr is not None
            return self._gr
        else:
            assert self._df is not None
            return self._df

    ###################################
    # adaptations of pyranges methods #
    ###################################

    EXCLUDED_PYRANGES_METHODS = [
        'lengths',
        'assign',
    ]

    def _pyranges_method_adapter_alone(methodname):
        def new_method(self, *args, **kwargs):
            result_gr = getattr(self.gr, methodname)(*args, **kwargs)
            if result_gr.empty:
                return self.__class__.init_empty(
                    refver=self.refver, 
                    annot_cols=self.annot_cols,
                )
            else:
                return self.spawn(result_gr)

        new_method.__name__ = methodname
        return new_method

    def _pyranges_method_adapter_other(methodname):
        def new_method(self, other, *args, **kwargs):
            if isinstance(other, pr.PyRanges):
                other_gr = other
            elif isinstance(other, GenomeDataFrame):
                other_gr = other.gr
            else:
                raise Exception(
                    f'"other" must be either PyRanges or GenomeDataFrame object. The type of "other" is: {type(other)}'
                )

            result_gr = getattr(self.gr, methodname)(other_gr, *args, **kwargs)

            if result_gr.empty:
                return self.__class__.init_empty(
                    refver=self.refver, 
                    annot_cols=self.annot_cols,
                )
            else:
                return self.spawn(result_gr)

        new_method.__name__ = methodname
        return new_method

    for key in set(dir(pr.PyRanges)).difference(EXCLUDED_PYRANGES_METHODS):
        if key.startswith('_'):
            continue
        mthd = getattr(pr.PyRanges, key)
        if not callable(mthd):
            continue
        
        params = inspect.signature(mthd).parameters
        if (
            (len(params) >= 2)
            and (list(params.keys())[1] == 'other')
        ):
            mthd = _pyranges_method_adapter_other(key)
            exec(f'{key} = mthd')
        else:
            mthd = _pyranges_method_adapter_alone(key)
            exec(f'{key} = mthd')

    ###########################################
    # adaptations of pandas DataFrame methods #
    ###########################################

    def __getitem__(self, key):
        return self.df.__getitem__(key).to_numpy().copy()

    def __setitem__(self, key, val):
        if self._df is None:
            self.set_df_from_gr()
            
        self._df.__setitem__(key, np.asarray(val))
        self.assign_df(self._df)

    @property
    def loc(self):
        return GenomeDataFrameLoc(self)

    @property
    def iloc(self):
        return GenomeDataFrameILoc(self)

    @staticmethod
    def dataframe_method_adaptor(methodname):
        original_method = getattr(pd.DataFrame, methodname)
        assert callable(original_method)

        sig = inspect.signature(original_method)
        has_inplace = ('inplace' in sig.parameters.keys())

        if has_inplace:
            def new_method(self, *args, **kwargs):
                ba = sig.bind(self, *args, **kwargs)
                ba.apply_defaults()
                inplace = ba.arguments['inplace']

                self.update_df()
                if inplace:
                    getattr(self._df, methodname)(*args, **kwargs)
                    self.assign_df(self._df)
                else:
                    new_df = getattr(self.df, methodname)(*args, **kwargs)
                    return self.spawn(new_df)

            new_method.__name__ = methodname
            return new_method
        else:
            # assume the original method returns a DataFrame
            def new_method(self, *args, **kwargs):
                new_df = getattr(self.df, methodname)(*args, **kwargs)
                return self.spawn(new_df)

        new_method.__name__ = methodname
        return new_method

    for mthdname in (
        'rename',
        'assign',
    ):
        new_mthd = dataframe_method_adaptor(mthdname)
        exec(f'{mthdname} = new_mthd')

    #################################
    # custom methods and properties #
    #################################

    @property
    def chromdict(self):
        return refgenome.get_chromdict(self.refver)

    ###############
    # coordinates #
    ###############

    @property
    def chromosomes(self):
        return self['Chromosome']

    chroms = chromosomes

    @property
    def chromosome_indexes(self):
        return np.fromiter(
            map(self.chromdict.contigs.index, self.chromosomes),
            dtype=int,
        )

    chromidxs = chromosome_indexes

    @property
    def starts(self):
        return self['Start']

    start0s = starts

    @property
    def ends(self):
        return self['End']

    end0s = ends

    def get_midpoints(self):
        return np.floor((self.start0s + self.end0s - 1) / 2).astype(int)

    @property
    def lengths(self):
        return self.end0s - self.start0s

    @property
    def lengths_cumsum(self):
        return np.cumsum(self.lengths)

    def check_self_overlap(self):
        if self.is_empty:
            return False
        else:
            bychrom = self.group_bychrom(how=None, sort=False)
            return any(
                genomedf_utils.check_interval_overlap(subgdf.start0s, subgdf.end0s)
                for chrom, subgdf in bychrom.items()
            )

    def check_duplicate_coords(self):
        if self.is_empty:
            return False
        else:
            return genomedf_utils.check_duplicate_coords(self.df)

    ##############################
    # basic dataframe attributes #
    ##############################

    @property
    def dtypes(self):
        return self.df.dtypes

    @property
    def shape(self):
        return self.df.shape

    @property
    def nrow(self):
        return self.df.shape[0]

    @property
    def ncol(self):
        return self.df.shape[1]

    @property
    def is_empty(self):
        newer_frame = self.get_newer_frame()
        if isinstance(newer_frame, pd.DataFrame):
            return newer_frame.shape[0] == 0
        else:
            return newer_frame.empty

    def copy(self):
        return self.spawn(self.df.copy())

    ######################
    # annotation columns #
    ######################

    @property
    def columns(self):
        if self.is_empty:
            return list(genomedf_utils.COMMON_COLUMNS)  # Chromosome, Start, End
        else:
            return self.get_newer_frame().columns.to_list()

    @property
    def annot_cols(self):
        return self.df.columns.drop(genomedf_utils.COMMON_COLUMNS).to_list()

    def choose_annots(self, annot_columns):
        annot_columns = np.atleast_1d(annot_columns)
        assert not set(annot_columns).intersection(genomedf_utils.COMMON_COLUMNS)
        return self.loc[:, genomedf_utils.COMMON_COLUMNS + list(annot_columns)]

    def drop_annots(self, columns=None, inplace=False):
        if columns is None:
            if inplace:
                self.assign_df(self.df.iloc[:, :3])
            else:
                return self.iloc[:, :3]
        else:
            columns = np.atleast_1d(columns)
            columns = set(columns).intersection(self.columns)
            #assert set(columns).issubset(self.columns)

            self_df = self.df
            if inplace:
                self_df.drop(columns, axis=1, inplace=True)
                self.assign_df(self_df)
            else:
                return self.spawn(self_df.drop(columns, axis=1, inplace=False))

    ##################
    # set operations #
    ##################

    def isec_union(self, other, nproc=1, drop_annots=False, sort=True):
        assert not self.check_self_overlap()
        assert not other.check_self_overlap()

        isec = self.intersect(other)
        self_diff_isec = self.subtract(isec)
        other_diff_isec = other.subtract(isec)
        if drop_annots:
            isec.drop_annots(inplace=True)
            self_diff_isec.drop_annots(inplace=True)
            other_diff_isec.drop_annots(inplace=True)
        else:
            cols_to_drop = ['Start_b', 'End_b', 'Distance']

            isec = isec.join(
                other, how='left', merge=None, find_nearest=False, nproc=nproc,
            )
            isec.drop_annots(cols_to_drop, inplace=True)

            self_diff_isec = self_diff_isec.join(
                other, how='left', find_nearest=True, merge=None, nproc=nproc,
            )
            self_diff_isec.drop_annots(cols_to_drop, inplace=True)

            other_diff_isec = other_diff_isec.join(
                self, how='left', find_nearest=True, merge=None, nproc=nproc,
            )
            other_diff_isec.drop_annots(cols_to_drop, inplace=True)

        result = self.__class__.concat([isec, self_diff_isec, other_diff_isec])
        if sort:
            result.sort()

        return result

    def subset_pyranges(self, chrom, start0=None, end0=None):
        """Returns a subset of original rows which overlaps the interval 
        specified by arguments. Original rows are returned as is, without 
        intersection with the input interval.
        'start0' and 'end0' arguments are 0-based & half-open.
        """
        getitem_key = (chrom, slice(start0, end0))
        return self.spawn(self.gr[getitem_key])

    def subset(self, chrom, start0=None, end0=None):
        """Returns a subset of original rows which overlaps the interval 
        specified by arguments. Original rows are returned as is, without 
        intersection with the input interval.
        'start0' and 'end0' arguments are 0-based & half-open.
        """
        isec_components = list()

        chrom_selector_idxs = self.get_chrom_selector(chrom, as_indexes=True)
        isec_components.append(chrom_selector_idxs)

        def helper(gdf_coords, query_pos, query_isend):
            relevant_gdf_coords = gdf_coords[chrom_selector_idxs]
            relevant_gdf_coords_selector = (
                (relevant_gdf_coords < query_pos)
                if query_isend else
                (relevant_gdf_coords > query_pos)
            )
            gdf_coords_selector_idxs = chrom_selector_idxs[relevant_gdf_coords_selector]
            return gdf_coords_selector_idxs

        if end0 is not None:
            isec_components.append(helper(self.start0s, end0, True))
        if start0 is not None:
            isec_components.append(helper(self.end0s, start0, False))

        final_selector_idxs = functools.reduce(np.intersect1d, isec_components)
        return self.iloc[final_selector_idxs, :]

    def subset_chroms(self, chromlist):
        return self.iloc[self.get_chrom_selector(chromlist, as_indexes=True), :]

    ########################
    # groupers & iterators #
    ########################

    def group_bychrom(self, how=None, sort=True):
        assert how in (None, 'gr', 'df')
        assert not self.is_empty

        result_asdf = dict(
            (chrom, subdf.reset_index(drop=True)) 
            for chrom, subdf in self.df.groupby(self.chroms, sort=False)
        )
        result_asgdf = {
            key: self.spawn(subdf) 
            for key, subdf in result_asdf.items()
        }
        if sort:
            for val in result_asgdf.values():
                val.sort()

        if how is None:
            return result_asgdf
        elif how == 'df':
            return {key: subgdf.df for key, subgdf in result_asgdf.items()}
        elif how == 'gr':
            return {key: subgdf.gr for key, subgdf in result_asgdf.items()}

    def get_chrom_selector(self, chromlist, as_indexes=False):
        chromlist = np.atleast_1d(chromlist)
        selector = np.isin(self.chromosomes, chromlist)
        if as_indexes:
            return np.nonzero(selector)[0]
        else:
            return selector

    def iter_coords(self):
        return zip(self.chromosomes, self.starts, self.ends)

    def get_coordcols_bychrom(self):
        self.sort()
        result = list()
        for (chrom, subdf) in self.df.iloc[:, :3].groupby(self.chroms, sort=False):
            #chroms = subdf['Chromosome'].to_numpy()
            start0s = subdf['Start'].to_numpy()
            end0s = subdf['End'].to_numpy()
            result.append((chrom, start0s, end0s))
        return result

    #def get_coord_groupkey(self):
    #    return genomedf_utils.get_coord_groupkey(self.df, self.chromdict)

    #########
    # split #
    #########

    @deco.get_deco_num_set_differently(('n', 'width'), 1)
    def equal_nrow_split(self, *, n=None, width=None):
        if n is not None:
            split_nums = tools.get_split_nums_byN(self.nrow, n)
        elif width is not None:
            split_nums = tools.get_split_nums_bywidth(self.nrow, width)

        indexes = np.insert(np.cumsum(split_nums), 0, 0)
        diff = np.diff(indexes)
        grouper = np.repeat(np.arange(len(diff)), diff)
        result = list(
            self.spawn(subdf) 
            for (key, subdf) in self.df.groupby(grouper, sort=False)
        )

        #result = list()
        #for start_idx, end_idx in tools.pairwise(indexes):
        #    result.append(self.iloc[start_idx:end_idx, :])

        return result

    @deco.get_deco_num_set_differently(('n', 'width'), 1)
    def equal_length_split(self, *, n=None, width=None):
        """Does NOT preserve original interval borders"""

        # get result_lengths_cumsum
        total_length = self.lengths_cumsum[-1]
        if n is not None:
            result_lengths_cumsum = np.cumsum(
                tools.get_split_nums_byN(total_length, n)
            )
        elif width is not None:
            result_lengths_cumsum = np.cumsum(
                tools.get_split_nums_bywidth(total_length, width)
            )

        # prepare result values
        result = list()
        for idx in range(len(result_lengths_cumsum)):
            # get start0 and end0 in the single merged coordinate system
            merged_start0 = (
                0 
                if idx == 0 else
                result_lengths_cumsum[idx - 1]
            )
            merged_end0 = result_lengths_cumsum[idx]
            # modify merged coordinates into interval-specific ones
            chrom_start, pos0_start, self_idx_start = self.equal_length_split_modify_coord(
                merged_start0, is_end=False,
            )
            chrom_end, pos0_end, self_idx_end = self.equal_length_split_modify_coord(
                merged_end0, is_end=True,
            )
            # create a GenomicDataFrame corresponding to a split unit
            gdf_data = {
                'chroms': list(),
                'start0s': list(),
                'end0s': list(),
            }

            if self_idx_start == self_idx_end:
                gdf_data['chroms'].append(chrom_start)
                gdf_data['start0s'].append(pos0_start)
                gdf_data['end0s'].append(pos0_end)
            else:
                gdf_data['chroms'].append(chrom_start)
                gdf_data['start0s'].append(pos0_start)
                gdf_data['end0s'].append(self.ends[self_idx_start])

                for self_idx in range(self_idx_start + 1, self_idx_end):
                    gdf_data['chroms'].append(self.chromosomes[self_idx])
                    gdf_data['start0s'].append(self.starts[self_idx])
                    gdf_data['end0s'].append(self.ends[self_idx])

                gdf_data['chroms'].append(chrom_end)
                gdf_data['start0s'].append(self.starts[self_idx_end])
                gdf_data['end0s'].append(pos0_end)

            result.append(self.spawn_from_data(**gdf_data))

        return result

    def equal_length_split_modify_coord(self, merged_pos0, is_end=False):
        lengths_cumsum = self.lengths_cumsum
        def get_interval_values(merged_pos0, idx):
            chrom = self.chromosomes[idx]

            if idx == 0:
                shift_within_interval = merged_pos0
            else:
                shift_within_interval = (
                    merged_pos0 
                    - lengths_cumsum[idx - 1]
                )
            new_pos0 = self.starts[idx] + shift_within_interval

            return chrom, new_pos0

        def handle_current_interval(
            merged_pos0, idx, length_previous, length_current, is_end,
        ):
            if (
                (merged_pos0 > length_previous)
                and (merged_pos0 <= length_current)
            ):
                if merged_pos0 == length_current:
                    if is_end:
                        chrom, new_pos0 = get_interval_values(merged_pos0, idx)
                    else:
                        chrom, new_pos0 = get_interval_values(merged_pos0, idx + 1)
                elif merged_pos0 < length_current:
                    chrom, new_pos0 = get_interval_values(merged_pos0, idx)

                to_break = True
            else:
                chrom = None
                new_pos0 = None
                to_break = False

            return chrom, new_pos0, to_break

        # sanity check
        assert merged_pos0 >= 0, f'"merged_pos0" must be non-negative.'
        assert not (
            (not is_end) 
            and (merged_pos0 >= lengths_cumsum[-1])
        ), (
            f'If ""is_end" is False, "merged_pos0" must be less than '
            f'the total IntervalList length.'
        )
        assert not (is_end and merged_pos0 > lengths_cumsum[-1]), (
            f'If ""is_end" is True, "merged_pos0" must be less than or '
            f'equal to the total IntervalList length.'
        )

        # main
        if merged_pos0 == 0:
            chrom = self.chromosomes[0]
            new_pos0 = self.starts[0]
            self_idx = 0
        else:
            for idx in range(len(lengths_cumsum)):
                length_current = lengths_cumsum[idx]
                length_previous = (
                    0
                    if idx == 0 else
                    lengths_cumsum[idx - 1]
                )
                chrom, new_pos0, to_break = handle_current_interval(
                    merged_pos0, idx, length_previous, length_current, is_end
                )
                self_idx = idx

                if to_break:
                    break

        return chrom, new_pos0, self_idx

    ################
    # segmentation #
    ################

    def make_compact(self):
        compact_pos0_list, decoder = genomedf_utils.compact_coords(self.start0s, self.end0s)
        compact_gdf = self.copy()
        compact_gdf['Start'] = compact_pos0_list
        compact_gdf['End'] = compact_gdf['Start'] + 1

        return compact_gdf, decoder

    @classmethod
    def get_segment_base(
        cls,
        gdf, 
        annot_colname,
        N_colname='N',
        smoothing=False, 
        smooth_kwargs=dict(), 
        segment_kwargs=dict(),
    ):
        # set kwargs
        segment_kwargs = (
            dict()
            | segment_kwargs
        )

        # make compact and run segmentation
        compact_gdf, decoder = gdf.make_compact()
        compact_seg_df = rdnacopy.run_dnacopy_segment(
            chromosomes=compact_gdf.chroms,
            positions=compact_gdf.start0s,
            values=compact_gdf[annot_colname],

            N_colname='N',
            value_colname='value',
            smoothing=smoothing, 
            verbose=False, 
            smooth_kwargs=smooth_kwargs, 
            segment_kwargs=segment_kwargs,
        )
        compact_seg_gdf = gdf.spawn(compact_seg_df)

        # recover noncompact coordinates
        noncompact_start0s, noncompact_end0s = genomedf_utils.uncompact_coords(
            compact_seg_gdf.starts, 
            compact_seg_gdf.ends, 
            decoder,
        )

        # make result
        result = GenomeDataFrame.from_data(
            refver=gdf.refver,
            **{
                'chroms': compact_seg_gdf.chromosomes,
                'start0s': noncompact_start0s,
                'end0s': noncompact_end0s,
                annot_colname: compact_seg_gdf['value'],
                N_colname: compact_seg_gdf['N'],
            }
        )
        return result

    def get_segment(
        self, 
        annot_colname,
        nproc=1,
        split_width=1000,

        N_colname='N',
        smoothing=False, 
        verbose=False, 
        smooth_kwargs=dict(), 
        segment_kwargs=dict(),
    ):
        assert not self.is_empty

        if verbose:
            logutils.log(f'Beginning segmentation')

        #grouper = self.chroms  # this may be more divided by cytoband features
        #grouped_gdfs = list(
        #    self.spawn(val) 
        #    for (key, val) in self.df.groupby(grouper, sort=False)
        #)

        self.sort()
        grouped_bychrom = self.group_bychrom(sort=False)
        grouped_gdfs = list()
        for chrom, subgdf in grouped_bychrom.items():
            grouped_gdfs.extend(subgdf.equal_nrow_split(width=split_width))

        with multiprocessing.Pool(nproc) as pool:
            args = (
                (
                    gdf, 
                    annot_colname,
                    N_colname,
                    smoothing,
                    smooth_kwargs,
                    segment_kwargs,
                )
                for gdf in grouped_gdfs
            )
            pool_result = pool.starmap(GenomeDataFrame.get_segment_base, args)

        result = GenomeDataFrame.concat(pool_result)
        result.sort()

        if verbose:
            logutils.log(f'Finished segmentation')

        return result

    ########
    # join #
    ########

    @deco.get_deco_arg_choices({'how': [None, 'left']})
    def join(
        self, other, 
        right_gdf_cols=None,
        how='left', 
        merge=['mean', 'std'],
        ddof=0,
        overlapping_length=False,
        N_colname='N',
        suffixes=None,
        merge_args=tuple(),
        merge_kwargs=dict(),
        winsorize=None,
        omit_N=False,
        find_nearest=False,

        split_width=3000,
        nproc=1,
    ):
        join_kwargs = dict(
            right_gdf_cols=right_gdf_cols,
            how=how,
            merge=merge,
            ddof=ddof,
            overlapping_length=overlapping_length,
            N_colname=N_colname,
            suffixes=suffixes,
            merge_args=merge_args,
            merge_kwargs=merge_kwargs,
            winsorize=winsorize,
            omit_N=omit_N,
            find_nearest=find_nearest,
        )

        if (
            (nproc == 1)
            or self.is_empty
            or other.is_empty
        ):
            return genomedf_methods.join_base(self, other, **join_kwargs)
        else:
            partial_left_gdf_list = genomedf_methods.split_left_gdf(self, width=split_width)
            with multiprocessing.Pool(nproc) as pool:
                args = (
                    (partial_left_gdf, other, join_kwargs)
                    for partial_left_gdf in partial_left_gdf_list
                )
                mp_result = pool.starmap(genomedf_methods.targetfunc, args)

            return self.__class__.concat(mp_result)

    ###########
    # binning #
    ###########

    def get_binsize(self):
        assert not self.is_empty

        self.sort()

        gdfs_bychrom = self.group_bychrom(how=None, sort=False)
        equal_length_portions = list()
        last_ones = list()
        for chrom, subgdf in gdfs_bychrom.items():
            lengths = subgdf.lengths
            last_ones.append(lengths[-1])
            if len(lengths) > 1:
                equal_length_portions.append(lengths[:-1])

        all_lengths = np.unique(np.concatenate(equal_length_portions))
        if all_lengths.shape != (1,):
            raise NotBinnedError(f'Bin lengths are not identical.')
        binsize = all_lengths[0]

        if len(last_ones) > 0:
            last_ones = np.asarray(last_ones)
            if not (last_ones <= binsize).all():
                raise NotBinnedError(f'Some last interval lengths are greater than the bin length.')

        return binsize

    def check_binned(self, return_binsize=False):
        try:
            binsize = self.get_binsize()
        except NotBinnedError:
            if return_binsize:
                return False, None
            else:
                return False
        else:
            if return_binsize:
                return True, binsize
            else:
                return True

    @staticmethod
    def get_upsizing_groupkey(singlechrom_df, chromlen, new_binsize, old_binsize):
        nrow = singlechrom_df.shape[0]
        assert nrow == np.ceil(chromlen / old_binsize).astype(int)

        n_newbin = np.ceil(chromlen / new_binsize).astype(int)
        olddf_idx_ends = np.fromiter(
            (
                min(int((new_binsize * x) / old_binsize), nrow)
                for x in range(1, n_newbin + 1)
            ),
            dtype=int,
        )
        group_lengths = np.diff(olddf_idx_ends, prepend=0)
        groupkey = np.repeat(range(len(group_lengths)), group_lengths)
        return groupkey, n_newbin

    @staticmethod
    def upsize_singlechrom_df(
        singlechrom_df, 
        new_binsize, 
        groupkey, 
        n_newbin, 
        chromlen, 
        annotcols_numeric,
        annotcols_bool,
        merge_bool,
    ):
        assert merge_bool in ('any', 'all')
        assert len(annotcols_numeric) > 0

        starts = np.arange(n_newbin) * new_binsize
        ends = np.concatenate([starts[1:], [chromlen]])
        result_common = pd.DataFrame(
            {
                'Chromosome': singlechrom_df['Chromosome'][0],
                'Start': starts,
                'End': ends,
            }
        )

        groupby = singlechrom_df.groupby(by=groupkey, sort=False)

        result_annot_mean = groupby[annotcols_numeric].mean()
        result_annot_mean.reset_index(inplace=True, drop=True)

        if len(annotcols_bool) > 0:
            if merge_bool == 'any':
                result_annot_any = groupby[annotcols_bool].any()
            elif merge_bool == 'all':
                result_annot_any = groupby[annotcols_bool].all()
            result_annot_any.reset_index(inplace=True, drop=True)

        if len(annotcols_bool) > 0:
            result_df = pd.concat([result_common, result_annot_mean, result_annot_any], axis=1)
        else:
            result_df = pd.concat([result_common, result_annot_mean], axis=1)

        return result_df

    def upsize_bin(self, size, merge_bool='any'):
        assert not self.is_empty

        old_binsize = self.get_binsize()
        if size <= old_binsize:
            raise Exception(f'New bin size ({size}) must be greater than the old bin size ({old_binsize})')

        dtypes = self.df.dtypes
        annotcols_numeric = list()
        annotcols_bool = list()
        for x in self.annot_cols:
            dtype = dtypes[x]
            if pd.api.types.is_bool_dtype(dtype):
                annotcols_bool.append(x)
            elif pd.api.types.is_numeric_dtype(dtype):
                annotcols_numeric.append(x)

        gdfs_bychrom = self.group_bychrom()
        newbin_dfs_bychrom = dict()
        for chrom, singlechrom_gdf in gdfs_bychrom.items():
            singlechrom_df = singlechrom_gdf.df
            chromlen = self.chromdict[chrom]
            groupkey, n_newbin = self.get_upsizing_groupkey(
                singlechrom_df=singlechrom_df, 
                chromlen=chromlen, 
                new_binsize=size, 
                old_binsize=old_binsize,
            )
            newbin_dfs_bychrom[chrom] = self.upsize_singlechrom_df(
                singlechrom_df=singlechrom_df, 
                new_binsize=size, 
                groupkey=groupkey, 
                n_newbin=n_newbin, 
                chromlen=chromlen, 
                annotcols_numeric=annotcols_numeric,
                annotcols_bool=annotcols_bool,
                merge_bool=merge_bool,
            )

        result_df = pd.concat(newbin_dfs_bychrom.values(), axis=0)
        result = self.spawn(result_df)
        return result

    ######################
    # other modification #
    ######################

    def sort(self):
        """In-place sort, only applied to df, not gr"""
        if self.is_empty:
            return

        if self._df is None:
            self.set_df_from_gr()
            do_sort = True
        elif self._gr is None:
            do_sort = (not self._sorted)
        else:
            if self.check_gr_is_newer() and (not self.check_df_gr_same()):
                self.set_df_from_gr()
                do_sort = True
            else:
                do_sort = (not self._sorted)

        if do_sort:
            self_df = self._df
            chrom_indexes = self_df['Chromosome'].apply(self.chromdict.contigs.index)
            sortkey = np.lexsort([self_df['End'], self_df['Start'], chrom_indexes])
            self.assign_df(self_df.iloc[sortkey, :])
            self._sorted = True

    def fill_gaps(self, edit_first_last=True):
        self.sort()

        new_start0s = list()
        new_end0s = list()
        for (chrom, old_start0s, old_end0s) in self.get_coordcols_bychrom():
            # fill borders between regions
            new_borders = np.floor(
                0.5 * (old_end0s[:-1] + old_start0s[1:])
            ).astype(int)
            current_new_start0s = np.insert(new_borders, 0, old_start0s[0])
            current_new_end0s = np.append(new_borders, old_end0s[-1])

            # modify start of the first region and end of the last region
            if edit_first_last:
                current_new_start0s[0] = 0
                current_new_end0s[-1] = self.chromdict[chrom]

            # apply
            new_start0s.append(current_new_start0s)
            new_end0s.append(current_new_end0s)

        new_start0s = np.concatenate(new_start0s)
        new_end0s = np.concatenate(new_end0s)

        result = self.spawn_from_data(
            chroms=self.chroms,
            start0s=new_start0s,
            end0s=new_end0s,
            **{key: self[key] for key in self.annot_cols},
        )
        return result


class GenomeDataFrameLoc:
    def __init__(self, gdf):
        self.gdf = gdf

    def __getitem__(self, key):
        result = self.gdf.df.loc[key]
        if isinstance(result, pd.DataFrame):
            return self.gdf.spawn(result)
        else:
            return result

    def __setitem__(self, key, val):
        self.gdf.df.loc.__setitem__(key, val)


class GenomeDataFrameILoc:
    def __init__(self, gdf):
        self.gdf = gdf

    def __getitem__(self, key):
        result = self.gdf.df.iloc[key]
        if isinstance(result, pd.DataFrame):
            return self.gdf.spawn(result)
        else:
            return result

    def __setitem__(self, key, val):
        self.gdf.df.iloc.__setitem__(key, val)





import functools
import multiprocessing

import pandas as pd
import numpy as np
import pyranges as pr

#from handygenome.refgenome.refgenome import NoXYError
from handygenome.genomedf.genomedf import GenomeDataFrame
import handygenome.cnv.cncall as cncall



class SegmentDataFrame(GenomeDataFrame):
    @staticmethod
    def filter_helper(values, cutoff):
        if cutoff is None:
            selector = True
        else:
            assert len(cutoff) == 2
            if cutoff[1] is None:
                selector1 = True
            else:
                selector1 = (values < cutoff[1])

            if cutoff[0] is None:
                selector2 = True
            else:
                selector2 = (values > cutoff[0])

            selector = np.logical_and(selector1, selector2)

        return selector

    #################################
    # merging of low ndata segments #
    #################################

    @staticmethod
    def merge_low_ndata_segments_targetfunc(subgdf, ndata_colname, cutoff=30):
        high_ndata_selector = (subgdf[ndata_colname] >= cutoff)
        high_ndata_selector_numeric = high_ndata_selector.astype(int)

        if high_ndata_selector.all():
            new_start0s = subgdf.start0s
            new_end0s = subgdf.end0s
            new_chroms = subgdf.chroms
        else:
            # make grouper
            cumsum = np.cumsum(high_ndata_selector_numeric)
            diff = np.diff(np.insert(high_ndata_selector_numeric, 0, 1))
            low_ndata_zone_starts = np.nonzero(diff == -1)[0]

            N = len(high_ndata_selector)
            offsets = functools.reduce(
                np.add, 
                [np.repeat([0, 1], [x, N - x]) for x in low_ndata_zone_starts]
            )
            grouper = cumsum + offsets

            # do groupby and filtering
            start_groupby = subgdf.df['Start'].groupby(grouper, sort=False)
            end_groupby = subgdf.df['End'].groupby(grouper, sort=False)

            new_start0s = start_groupby.first().to_numpy()
            new_end0s = end_groupby.last().to_numpy()
            new_chroms = np.repeat(subgdf.chroms[0], len(new_end0s))

        return new_chroms, new_start0s, new_end0s

    def merge_low_ndata_segments(self, ndata_colname, cutoff=30, nproc=1):
        assert ndata_colname in self.columns

        bychrom = self.group_bychrom(sort=True)
        args = ((subgdf, ndata_colname, cutoff) for subgdf in bychrom.values())
        with multiprocessing.Pool(nproc) as pool:
            mp_result = pool.starmap(
                self.__class__.merge_low_ndata_segments_targetfunc, 
                args,
            )

        result_chroms, result_start0s, result_end0s = zip(*mp_result)
        result_chroms = np.concatenate(result_chroms)
        result_start0s = np.concatenate(result_start0s)
        result_end0s = np.concatenate(result_end0s)

        result = self.spawn_from_data(
            chroms=result_chroms,
            start0s=result_start0s,
            end0s=result_end0s,
        )
        return result

    @staticmethod
    def incoporate_low_ndata_segments_targetfunc(subgdf, ndata_colname, chrom_length, cutoff=30):
        start0s = subgdf.start0s
        end0s = subgdf.end0s
        ndata_array = subgdf[ndata_colname]
        high_ndata_selector = (ndata_array >= cutoff)
        low_ndata_selector = np.logical_not(high_ndata_selector)

        # 1. remove low ndata segments at borders
        if low_ndata_selector[0]:  # the first segment is with low ndata
            start0s = np.delete(start0s, 0)
            end0s = np.delete(end0s, 0)
            ndata_array = np.delete(ndata_array, 0)
            high_ndata_selector = np.delete(high_ndata_selector, 0)
            low_ndata_selector = np.delete(low_ndata_selector, 0)

            start0s[0] = 0

        if low_ndata_selector[-1]:  # the last segment is with low ndata
            start0s = np.delete(start0s, -1)
            end0s = np.delete(end0s, -1)
            ndata_array = np.delete(ndata_array, -1)
            high_ndata_selector = np.delete(high_ndata_selector, -1)
            low_ndata_selector = np.delete(low_ndata_selector, -1)

            #end0s[-1] = self.chromdict[subgdf.chroms[0]]
            end0s[-1] = chrom_length

        # 2. handle intervening low ndata segment
        if low_ndata_selector.any():
            low_ndata_indexes = np.nonzero(low_ndata_selector)[0]
            low_ndata_midpoints = np.floor(
                (start0s[low_ndata_indexes] + end0s[low_ndata_indexes] - 1) / 2
            ).astype(int)

            before_indexes = low_ndata_indexes - 1
            end0s[before_indexes] = low_ndata_midpoints

            after_indexes = low_ndata_indexes + 1
            start0s[after_indexes] = low_ndata_midpoints

        # 3. result
        new_start0s = start0s[high_ndata_selector]
        new_end0s = end0s[high_ndata_selector]
        new_chroms = np.repeat(subgdf.chroms[0], len(new_start0s))

        return new_chroms, new_start0s, new_end0s

    def incoporate_low_ndata_segments(self, ndata_colname, cutoff=30, nproc=1):
        assert ndata_colname in self.columns

        bychrom = self.group_bychrom(sort=True)
        subgdf_list = list(bychrom.values())
        chrom_length_list = [self.chromdict[x.chroms[0]] for x in subgdf_list]
        args = (
            (subgdf, ndata_colname, chrom_length, cutoff)
            for (subgdf, chrom_length) in zip(subgdf_list, chrom_length_list)
        )
        with multiprocessing.Pool(nproc) as pool:
            mp_result = pool.starmap(
                self.__class__.incoporate_low_ndata_segments_targetfunc, 
                args,
            )

        result_chroms, result_start0s, result_end0s = zip(*mp_result)
        result_chroms = np.concatenate(result_chroms)
        result_start0s = np.concatenate(result_start0s)
        result_end0s = np.concatenate(result_end0s)

        result = self.spawn_from_data(
            chroms=result_chroms,
            start0s=result_start0s,
            end0s=result_end0s,
        )
        return result

    ###############
    # default CNg #
    ###############

    def add_default_CNg_Bg(self, is_female):
        if cncall.DEFAULT_CNG_COLNAME not in self.columns:
            CNg_Bg_gdf = cncall.get_default_CNg_Bg_diploid(self.refver, is_female)
            annot_seg = self.join(
                CNg_Bg_gdf, 
                how='left',
                right_gdf_cols=[cncall.DEFAULT_CNG_COLNAME, cncall.DEFAULT_BG_COLNAME],
                merge='longest',
                overlapping_length=True,
                omit_N=True,
                suffixes={'longest': ''},
            )
            self.assign_df(annot_seg.df)

    def get_default_CNg(self, is_female):
        if cncall.DEFAULT_CNG_COLNAME not in self.columns:
            self.add_default_CNg_Bg(is_female)
        return self[cncall.DEFAULT_CNG_COLNAME]

    def get_default_Bg(self, is_female):
        if cncall.DEFAULT_BG_COLNAME not in self.columns:
            self.add_default_CNg_Bg(is_female)
        return self[cncall.DEFAULT_BG_COLNAME]





import textwrap

import handygenome.workflow as workflow
import handygenome.vcfeditor.split as split_module
import handygenome.vcfeditor.indexing as indexing


def argument_parser(cmdargs):
    def sanity_check(args):
        if (args.n_line, args.n_file).count(None) != 1:
            e_msg = f"Exactly one of '--nfile' or '--nline' must be used."
            raise SystemExit(e_msg)  # this does not show traceback

    parser_dict = workflow.init_parser(description=textwrap.dedent("""\
        - Splits input vcf by either 'nfile' (number of output files) or 
          'nline' (number of variant record lines each split file contains).
        - Split file names looks like: 
          <prefix>000<suffix>, <prefix>001<suffix>, ...
        - Exactly one of '--nfile' or '--nline' options must be used.
        """))

    workflow.add_infile_arg(parser_dict['required'], required=True)
    workflow.add_outdir_arg(parser_dict['required'], required=True)
    workflow.add_outfmt_arg(parser_dict['optional'], required=False, 
                            default='z')
    workflow.add_index_arg(parser_dict)

    parser_dict['optional'].add_argument(
        '--nfile', dest='n_file', required=False, type=int,
        help=textwrap.dedent(f"""\
            The number of output files. 
            Must not be used at the same time with --nline option."""))
    parser_dict['optional'].add_argument(
        '--nline', dest='n_line', required=False, type=int,
        help=textwrap.dedent(f"""\
            The number of variant record lines per output file.
            Must not be used at the same time with --nfile option."""))
    parser_dict['optional'].add_argument(
        '--prefix', dest='prefix', required=False, default='',
        help=textwrap.dedent(f"""\
            prefix for the names of split files. Default is an empty
            string."""))
    parser_dict['optional'].add_argument(
        '--suffix', dest='suffix', required=False, default='.vcf.gz',
        help=f'suffix for the names of split files.')

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def main(cmdargs):
    args = argument_parser(cmdargs)
    split_module.main(vcf_path=args.infile_path, outdir=args.outdir_path,
                      n_file=args.n_file, n_line=args.n_line,
                      mode_pysam=args.mode_pysam, prefix=args.prefix,
                      suffix=args.suffix)

    if not args.donot_index:
        indexing.index_vcf(args.outfile_path)



import os
import sys
import re
import subprocess
import multiprocessing
import tempfile

import sequenza.commands

import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))


def 


import multiprocessing
import functools

import pandas as pd
import numpy as np
import pyranges as pr
import pysam

import handygenome.peakutils as peakutils
import handygenome.logutils as logutils
from handygenome.genomedf.genomedf import GenomeDataFrame
import handygenome.cnv.cncall as cncall
from handygenome.cnv.segment import SegmentDataFrame
from handygenome.cnv.depth import DepthSegmentDataFrame
from handygenome.cnv.baf import BAFSegmentDataFrame


CLONAL_CN_COLNAME = 'CN'
CLONAL_B_COLNAME = 'B'
SUBCLONAL_CN_COLNAME = 'subCN'
SUBCLONAL_B_COLNAME = 'subB'
CCF_COLNAME = 'ccf'


#class CNVSegmentDataFrame(SegmentDataFrame):
class CNVSegmentDataFrame(DepthSegmentDataFrame, BAFSegmentDataFrame):
    @classmethod
    def from_segments(cls, seg_gdfs, nproc=1):
        assert len(seg_gdfs) > 0

        refvers = set(x.refver for x in seg_gdfs)
        if len(refvers) != 1:
            raise Exception(f'refver differs between segment gdfs: {refvers}')
        refver = refvers.pop()

        merged_seg_gdf = functools.reduce(
            lambda seg1, seg2: (
                seg1.drop_annots('N', inplace=False).isec_union(
                    seg2.drop_annots('N', inplace=False), 
                    nproc=nproc,
                    drop_annots=False,
                    sort=True,
                )
            ),
            seg_gdfs,
        )

        return cls.from_frame(merged_seg_gdf.df, refver=refver)

    # assign

    def assign_clonal_CN(self, data):
        self[CLONAL_CN_COLNAME] = data

    def assign_clonal_B(self, data):
        self[CLONAL_B_COLNAME] = data

    def assign_subclonal_CN(self, data):
        self[SUBCLONAL_CN_COLNAME] = data

    def assign_subclonal_B(self, data):
        self[SUBCLONAL_B_COLNAME] = data

    def assign_ccf(self, data):
        self[CCF_COLNAME] = data

    # fetch

    def get_clonal_CN(self):
        return self[CLONAL_CN_COLNAME]

    def get_clonal_B(self):
        return self[CLONAL_B_COLNAME]

    def get_subclonal_CN(self):
        return self[SUBCLONAL_CN_COLNAME]

    def get_subclonal_B(self):
        return self[SUBCLONAL_B_COLNAME]

    def get_ccf(self):
        return self[CCF_COLNAME]





import os
import warnings
import re
import logging
import copy
import pprint
import collections
import functools

import pysam
import numpy as np
import scipy.stats

import handygenome.logutils as logutils
import handygenome.workflow as workflow
import handygenome.annotation.annotitem as annotitem
import handygenome.variant.infoformat as infoformat
import handygenome.read.readplus as readplus
import handygenome.read.alleleinfo as liballeleinfo


ZERO_ONE_UNIFORM = scipy.stats.uniform(loc=0, scale=1)
DEFAULT_DEPTH_LIMITS = (0, -1)
DEFAULT_MQ_LIMITS = (0, -1)


class AlleleclassError(Exception):
    pass


def get_position_info(bam, chrom, pos0):
    mqlist = list()
    idx = -1
    for idx, read in enumerate(bam.fetch(chrom, pos0, pos0 + 1)):
        mqlist.append(read.mapping_quality)
    depth = idx + 1
    return depth, mqlist


def get_position_info_pileup(bam, chrom, pos0):
    pup = bam.pileup(
        chrom, pos0, pos0 + 1, 
        truncate=True, 
        stepper='nofilter', 
        ignore_overlaps=False, 
        flag_filter=0, 
        ignore_orphans=False, 
        min_base_quality=0, 
        min_mapping_quality=0,
    )
    pupcol = next(pup)
    mqlist = pupcol.get_mapping_qualities()
    depth = pupcol.get_num_aligned()
    return depth, mqlist


class ReadStats(annotitem.AnnotItemFormatSingle):
    @staticmethod
    def handle_limits_arg(limits):
        result = list(limits)
        if result[1] == -1:
            result[1] = np.inf
        return result

    @staticmethod
    def check_position_validity(depth, mqlist, depth_limits, mq_limits):
        mean_mq = np.mean(mqlist) if mqlist else np.nan
        depth_okay = (depth >= depth_limits[0]) and (depth <= depth_limits[1])
        mq_okay = (
            np.isnan(mean_mq)
            or ((mean_mq >= mq_limits[0]) and (mean_mq <= mq_limits[1]))
        )
        return depth_okay and mq_okay
        
    @classmethod
    def from_bam(
        cls, 
        vcfspec, 
        bam, 
        fasta=None, 
        chromdict=None,
        rpplist_kwargs=dict(),
        alleleinfo_kwargs=dict(),
        countonly=False,
        depth_limits=DEFAULT_DEPTH_LIMITS,
        mq_limits=DEFAULT_MQ_LIMITS,
        include_mNM_items=False,
    ):
        depth, mqlist = get_position_info(bam, vcfspec.chrom, vcfspec.pos0)
        depth_limits = cls.handle_limits_arg(depth_limits)
        mq_limits = cls.handle_limits_arg(mq_limits)

        if cls.check_position_validity(depth, mqlist, depth_limits, mq_limits):
            readstats_data = get_readstats_data(
                vcfspec, bam, fasta, chromdict,
                rpplist_kwargs=rpplist_kwargs,
                alleleinfo_kwargs=alleleinfo_kwargs,
                countonly=countonly,
            )
            result = cls.from_readstats_data(
                readstats_data, 
                vcfspec, 
                fasta, 
                chromdict,
                countonly=countonly,
                include_mNM_items=include_mNM_items,
            )
            del readstats_data
        else:
            result = cls.init_invalid(vcfspec, fasta, chromdict, countonly=countonly)

        return result

    @classmethod
    def from_readstats_data(
        cls, 
        readstats_data,  
        vcfspec, 
        fasta=None, 
        chromdict=None,
        countonly=False,
        include_mNM_items=False,
    ):
        def allele_means(data):
            return dict(
                (
                    alleleclass, 
                    np.nan if len(values) == 0 else np.mean(values),
                )
                for (alleleclass, values) in data.items()
            )

        def allele_medians(data):
            return dict(
                (
                    alleleclass, 
                    np.nan if len(values) == 0 else np.median(values),
                ) 
                for (alleleclass, values) in data.items()
            )

        def varpos_kstest(data):
            summary = dict()
            for alleleclass, values in data.items():
                if len(values) == 0:
                    pval = np.nan
                else:
                    pval = scipy.stats.kstest(values, ZERO_ONE_UNIFORM.cdf).pvalue
                summary[alleleclass] = pval

            return summary

        def handle_orient(data, mode):
            summary = dict()
            for alleleclass, values in data.items():
                if len(values) == 0:
                    summary[alleleclass] = np.nan
                else:
                    counter = collections.Counter(values)
                    if mode == 'pairorient':
                        x = (counter['f1r2'], counter['f2r1'])
                    elif mode == 'readorient':
                        x = (counter['f'], counter['r'])
                    summary[alleleclass] = scipy.stats.binom_test(x, p=0.5, alternative='two-sided')

            return summary

        def handle_mNM(
            data, 
            rppcounts_dict, 
            rpp_range0_dict, 
            recurrence_cutoff_fraction,
            recurrence_cutoff_denominator,
        ):
            summary_without_recurrence = dict()
            summary_only_recurrence = dict()
            mNMitems_norecur = dict()
            mNMitems_recur = dict()
            for alleleclass, values in data.items():    
                # set values
                mNM_count_sum_with_recurrence = 0
                mNM_count_sum_without_recurrence = 0

                mNMitems_norecur[alleleclass] = list()
                mNMitems_recur[alleleclass] = list()

                if len(values) > 0:
                    counter = collections.Counter(values)
                    for key, val in counter.items():
                        refpos0 = key[0]
                        relevant_rppcount = sum(
                            any((refpos0 in x) for x in range0_list)
                            for range0_list in rpp_range0_dict[alleleclass]
                        )
                        if relevant_rppcount == 0:
                            raise Exception(
                                f'The number of relevant rpp is 0. mNM element: {key}'
                            )

                        if (
                            (val / relevant_rppcount) >= recurrence_cutoff_fraction and
                            relevant_rppcount >= recurrence_cutoff_denominator
                        ):  # This means a recurrent MM
                            mNM_count_sum_with_recurrence += 1
                            mNMitems_recur[alleleclass].append(key)
                        else:
                            mNM_count_sum_without_recurrence += 1
                            mNMitems_norecur[alleleclass].append(key)

                # summarize
                if rppcounts_dict[alleleclass] == 0:
                    summary_without_recurrence[alleleclass] = np.nan
                else:
                    summary_without_recurrence[alleleclass] = (
                        mNM_count_sum_without_recurrence / rppcounts_dict[alleleclass]
                    )
                summary_only_recurrence[alleleclass] = mNM_count_sum_with_recurrence

            return summary_without_recurrence, summary_only_recurrence, mNMitems_norecur, mNMitems_recur

        # main
        result = cls(is_missing=False)
        result.is_invalid = False
        result.vcfspec = vcfspec
        result.fasta = (
            refgenome.get_fasta(vcfspec.refver)
            if fasta is None else
            fasta
        )
        result.chromdict = (
            refgenome.get_chromdict(vcfspec.refver)
            if chromdict is None else
            chromdict
        )

        result['rppcounts'] = readstats_data['count'].copy()

        if not countonly:
            result['mean_BQs'] = allele_means(readstats_data['BQ'])
            result['median_BQs'] = allele_medians(readstats_data['BQ'])

            result['mean_MQs'] = allele_means(readstats_data['MQ'])
            result['median_MQs'] = allele_medians(readstats_data['MQ'])

            result['mean_cliplens'] = allele_means(readstats_data['cliplen'])
            result['median_cliplens'] = allele_medians(readstats_data['cliplen'])

            result['mNM'], result['recurrent_mNM'], mNMitems_norecur, mNMitems_recur = handle_mNM(
                readstats_data['mNM'], 
                readstats_data['count'], 
                readstats_data['rpp_range0'], 
                recurrence_cutoff_fraction=0.9,
                recurrence_cutoff_denominator=3,
            )
            if include_mNM_items:
                result['mNM_items'] = mNMitems_norecur
                result['recurrent_mNM_items'] = mNMitems_recur

            result['pairorient_pvalues'] = handle_orient(readstats_data['pairorient'], mode='pairorient')
            result['readorient_pvalues'] = handle_orient(readstats_data['readorient'], mode='pairorient')

            result['varpos_uniform_pvalues'] = varpos_kstest(readstats_data['pos0_left_fraction'])
            result['mean_varpos_fractions'] = allele_means(readstats_data['pos0_left_fraction'])
            result['median_varpos_fractions'] = allele_medians(readstats_data['pos0_left_fraction'])

        return result

    @classmethod
    def init_invalid(cls, vcfspec, fasta=None, chromdict=None, countonly=False, verbose=True):
        if verbose:
            logutils.print_timestamp(f'Initiating ReadStats object as invalid mode')
        # main
        result = cls(is_missing=False)
        result.is_invalid = True
        result.vcfspec = vcfspec
        result.fasta = (
            refgenome.get_fasta(vcfspec.refver)
            if fasta is None else
            fasta
        )
        result.chromdict = (
            refgenome.get_chromdict(vcfspec.refver)
            if chromdict is None else
            chromdict
        )

        result['rppcounts'] = {
            alleleclass: np.nan 
            for alleleclass in vcfspec.get_alleleclasses()
        }
        result['rppcounts']['softclip_overlap'] = np.nan

        if not countonly:
            for key in (
                'mean_BQs',
                'median_BQs',
                'mean_MQs',
                'median_MQs',
                'mean_cliplens',
                'median_cliplens',
                'mNM',
                'recurrent_mNM',
                'pairorient_pvalues',
                'readorient_pvalues',
                'varpos_uniform_pvalues',
                'mean_varpos_fractions',
                'median_varpos_fractions',
            ):
                result[key] = {
                    alleleclass: np.nan 
                    for alleleclass in vcfspec.get_alleleclasses()
                }

        return result

    #def write(self, vr, sampleid):
    #    return self.write_base(vr, sampleid)

    def get_allele_indexes_average(self, key, allele_indexes=None):
        assert key != 'rppcounts'
        if allele_indexes is None:
            allele_indexes = tuple(self[key].keys())

        values = np.fromiter((self[key][x] for x in allele_indexes), dtype=float)
        weights = np.fromiter((self['rppcounts'][x] for x in allele_indexes), dtype=float)

        nan_indexes = np.logical_or(np.isnan(values), np.isnan(weights))
        values = values[~nan_indexes]
        weights = weights[~nan_indexes]

        if sum(weights) == 0:
            return np.nan
        else:
            return np.average(values, weights=weights)

    get_allele_indexes_mean = get_allele_indexes_average
    #get_alleleindexes_mean = get_allele_indexes_average

    def get_total_rppcount(self, exclude_other=False):
        if exclude_other:
            return sum(
                val for (key, val) in self['rppcounts'].items()
                if isinstance(key, int) and key != -1
            )
        else:
            return sum(
                val for (key, val) in self['rppcounts'].items()
                if isinstance(key, int)
            )

    def get_vaf(self, alleleclass=1, exclude_other=False):
        if (alleleclass == -1) and exclude_other:
            raise Exception(f'When "alleleclass" is -1, "exclude_other" must be False.')

        total_rppcount = self.get_total_rppcount(exclude_other=exclude_other)
        if total_rppcount == 0:
            return np.nan
        else:
            try:
                return self['rppcounts'][alleleclass] / total_rppcount
            except KeyError:
                raise AlleleclassError(f'Invalid alleleclass')

    def get_sorted_vafs(self, exclude_other=True, reverse=True):
        vafs = list()
        for alleleclass in self['rppcounts'].keys():
            if alleleclass is None:
                continue
            if exclude_other:
                if alleleclass == -1:
                    continue
            vafs.append(
                self.get_vaf(alleleclass=alleleclass, exclude_other=exclude_other)
            )

        return sorted(vafs, reverse=reverse)

    #def get_all_allele_rppcounts(self):
    #    result = dict()
        #total_rppcount = self.get_total_rppcount()
    #    for alleleclass, rppcount in self['rppcounts'].items():
    #        result[alleleclass] = rppcount / total_rppcount
    #    return result


class ReadStatsSampledict(annotitem.AnnotItemFormatSampledict):
    meta = {
        'ID': 'readstats', 'Number': 1, 'Type': 'String',
        'Description': 'Read-based statistics for the variant.',
    }
    unit_class = ReadStats

    @staticmethod
    def handle_limits_arg(limits, bam_dict):
        if isinstance(limits, (tuple, list)):
            new_limits = {key: list(limits) for key in bam_dict.keys()}
        elif isinstance(limits, dict):
            #if set(limits.keys()) != set(bam_dict.keys()):
            if not set(bam_dict.keys()).issubset(set(limits.keys())):
                raise Exception(f'Keys of "limits" argument are not a superset of the keys of "bam_dict".')
            new_limits = limits
        else:
            raise Exception(f'"limits" argument must be either a tuple, list, or dict')

        return new_limits

    @classmethod
    def from_bam_dict(
        cls, 
        bam_dict, 
        vcfspec, 
        fasta=None, 
        chromdict=None,
        rpplist_kwargs=dict(),
        alleleinfo_kwargs=dict(),
        countonly=False,
        depth_limits=DEFAULT_DEPTH_LIMITS,
        mq_limits=DEFAULT_MQ_LIMITS,
        init_invalid=False,
        include_mNM_items=False,
    ):
        depth_limits = cls.handle_limits_arg(depth_limits, bam_dict)
        mq_limits = cls.handle_limits_arg(mq_limits, bam_dict)

        result = cls()
        if init_invalid:
            for sampleid, bam in bam_dict.items():
                result[sampleid] = ReadStats.init_invalid(vcfspec, fasta, chromdict, countonly=countonly)
        else:
            for sampleid, bam in bam_dict.items():
                result[sampleid] = ReadStats.from_bam(
                    vcfspec, bam, fasta, chromdict,
                    rpplist_kwargs=rpplist_kwargs,
                    alleleinfo_kwargs=alleleinfo_kwargs,
                    countonly=countonly,
                    depth_limits=depth_limits[sampleid],
                    mq_limits=mq_limits[sampleid],
                    include_mNM_items=include_mNM_items,
                )
        return result

    @classmethod
    def from_vr(cls, vr, sampleid_list=None):
        return cls.from_vr_base(vr, sampleid_list=sampleid_list)

    def get_first_readstats(self):
        return next(iter(self.values()))

    @property
    def vcfspec(self):
        return self.get_first_readstats().vcfspec

    @property
    def fasta(self):
        return self.get_first_readstats().fasta

    @property
    def chromdict(self):
        return self.get_first_readstats().chromdict

    def update_bams(
        self, 
        bam_dict,
        rpplist_kwargs=dict(),
        alleleinfo_kwargs=dict(),
        countonly=False,
        depth_limits=DEFAULT_DEPTH_LIMITS,
        mq_limits=DEFAULT_MQ_LIMITS,
        init_invalid=False,
        include_mNM_items=False,
    ):
        if len(self) == 0:
            raise Exception(f'Length of self must be greater than 0')

        depth_limits = self.handle_limits_arg(depth_limits, bam_dict)
        mq_limits = self.handle_limits_arg(mq_limits, bam_dict)

        if init_invalid:
            for sampleid, bam in bam_dict.items():
                self[sampleid] = ReadStats.init_invalid(
                    self.vcfspec, self.fasta, self.chromdict, countonly=countonly,
                )
        else:
            for sampleid, bam in bam_dict.items():
                self[sampleid] = ReadStats.from_bam(
                    self.vcfspec, bam, self.fasta, self.chromdict,
                    rpplist_kwargs=rpplist_kwargs,
                    alleleinfo_kwargs=alleleinfo_kwargs,
                    countonly=countonly,
                    depth_limits=depth_limits[sampleid],
                    mq_limits=mq_limits[sampleid],
                    include_mNM_items=include_mNM_items,
                )

    def write(self, vr, donot_write_missing=True):
        return self.write_base(vr, donot_write_missing=donot_write_missing)

    def show_rppcounts(self):
        for sampleid, readstats in self.items():
            print(sampleid)
            pprint.pprint(readstats['rppcounts'])

    def show_vafs(self):
        for sampleid, readstats in self.items():
            print(sampleid)
            vaf_dict = {
                alleleclass: readstats.get_vaf(alleleclass)
                for alleleclass in sorted(
                    readstats['rppcounts'].keys(), 
                    key=(
                        lambda x: (
                            -2 if x is None else 
                            (np.inf if x == 'softclip_overlap' else x)
                        )
                    )
                )
            }
            pprint.pprint(vaf_dict)


def rpplist_to_readstats_data(
    rpplist, vcfspec, flanklen=liballeleinfo.DEFAULT_FLANKLEN,
):
    def add_var_pos0s_rp(rp, data, alleleclass_rpp, vcfspec):
        var_querypos0s = rp.get_querypos0_of_range0_allmodes(vcfspec.REF_range0)

        if var_querypos0s['left'] is not None:
            data['pos0_left'][alleleclass_rpp].append(var_querypos0s['left'])
            data['pos0_right'][alleleclass_rpp].append(var_querypos0s['right'])
            data['pos0_5prime'][alleleclass_rpp].append(var_querypos0s['5prime'])
            data['pos0_3prime'][alleleclass_rpp].append(var_querypos0s['3prime'])

            data['pos0_left_fraction'][alleleclass_rpp].append(var_querypos0s['left_fraction'])
            data['pos0_right_fraction'][alleleclass_rpp].append(var_querypos0s['right_fraction'])
            data['pos0_5prime_fraction'][alleleclass_rpp].append(var_querypos0s['5prime_fraction'])
            data['pos0_3prime_fraction'][alleleclass_rpp].append(var_querypos0s['3prime_fraction'])

    # initialize
    alleleclass_keys = vcfspec.get_alleleclasses()
    data = dict()

    # fields initialized as integer
    data['count'] = {x: 0 for x in alleleclass_keys}
    data['count']['softclip_overlap'] = 0

    # fields initialized as list
    for key in (
        'rpp_range0',
        'MQ', 'BQ', 'mNM', 'clipspec', 'cliplen', 'pairorient', 'readorient',
        'pos0_left', 'pos0_right', 'pos0_5prime', 'pos0_3prime',
        'pos0_left_fraction', 'pos0_right_fraction', 
        'pos0_5prime_fraction', 'pos0_3prime_fraction',
    ):
        data[key] = {x: list() for x in alleleclass_keys}

    # add data
    for rpp in rpplist:
        if vcfspec not in rpp.alleleclass.keys():
            rpp.update_alleleclass(vcfspec=vcfspec, flanklen=flanklen)
        alleleclass_rpp = rpp.alleleclass[vcfspec]

        # count
        data['count'][alleleclass_rpp] += 1
        if alleleclass_rpp is None:
            if rpp.check_softclip_overlaps_vcfspec(vcfspec):
                data['count']['softclip_overlap'] += 1

        # rpp_range0
        rpp_range0s = list()
        if rpp.rp1.read.reference_name == vcfspec.chrom:
            rpp_range0s.append(rpp.rp1.range0)
        if rpp.rp2 is not None:
            if rpp.rp2.read.reference_name == vcfspec.chrom:
                rpp_range0s.append(rpp.rp2.range0)
        data['rpp_range0'][alleleclass_rpp].append(rpp_range0s)

        # MQ
        data['MQ'][alleleclass_rpp].append(rpp.get_MQ())

        # BQ
        data['BQ'][alleleclass_rpp].extend(rpp.rp1.get_BQlist(vcfspec))
        if rpp.rp2 is not None:
            data['BQ'][alleleclass_rpp].extend(rpp.rp2.get_BQlist(vcfspec))

        # cliplen
        data['cliplen'][alleleclass_rpp].append(rpp.get_cliplen())

        # pairorient
        data['pairorient'][alleleclass_rpp].append(rpp.pairorient)

        # readorient
        if rpp.rp1.alleleclass[vcfspec] == alleleclass_rpp:
            data['readorient'][alleleclass_rpp].append(
                'f' if rpp.rp1.read.is_forward else 'r'
            )
        if rpp.rp2 is not None:
            if rpp.rp2.alleleclass[vcfspec] == alleleclass_rpp:
                data['readorient'][alleleclass_rpp].append(
                    'f' if rpp.rp2.read.is_forward else 'r'
                )

        # mNM
        mNM_data, clipspec_data = rpp.get_mNM_clipspec_data(vcfspec)
        data['mNM'][alleleclass_rpp].extend(mNM_data)
        data['clipspec'][alleleclass_rpp].extend(clipspec_data)

        # var_pos0s
        add_var_pos0s_rp(rpp.rp1, data, alleleclass_rpp, vcfspec)
        if rpp.rp2 is not None:
            add_var_pos0s_rp(rpp.rp2, data, alleleclass_rpp, vcfspec)

    # turn into numpy arrays
#    for key in data.keys():
#        if key != 'count':
#            data[key] = {
#                subkey: np.array(subval)
#                for subkey, subval in data[key].items()
#            }

    return data


def rpplist_to_readstats_data_countonly(
    rpplist, vcfspec, flanklen=liballeleinfo.DEFAULT_FLANKLEN,
):
    # initialize
    alleleclass_keys = vcfspec.get_alleleclasses()
    data = dict()
    data['count'] = {x: 0 for x in alleleclass_keys}
    data['count']['softclip_overlap'] = 0
    # add data
    for rpp in rpplist:
        if vcfspec not in rpp.alleleclass.keys():
            rpp.update_alleleclass(vcfspec=vcfspec, flanklen=flanklen)
        alleleclass_rpp = rpp.alleleclass[vcfspec]
        data['count'][alleleclass_rpp] += 1

    return data


def get_readstats_data(
    vcfspec, bam, 
    fasta=None, 
    chromdict=None, 
    rpplist_kwargs=dict(),
    alleleinfo_kwargs=dict(),
    countonly=False,
):
    """Only for non-sv cases"""

    #rpplist_kwargs.update({'view': False})
    rpplist_kwargs['view'] = False
    rpplist = readplus.get_rpplist_nonsv(
        bam=bam, 
        fasta=fasta, 
        chromdict=chromdict, 
        chrom=vcfspec.chrom, 
        start0=vcfspec.pos0, 
        end0=vcfspec.end0, 
        **rpplist_kwargs,
    )
    rpplist.update_alleleclass(
        vcfspec=vcfspec, 
        **alleleinfo_kwargs,
    )
    if countonly:
        readstats_data = rpplist_to_readstats_data_countonly(rpplist, vcfspec)
    else:
        readstats_data = rpplist_to_readstats_data(rpplist, vcfspec)

    del rpplist

    return readstats_data





#!/usr/bin/env python3
 
#Author: Erik Bergstrom

#Contact: ebergstr@eng.ucsd.edu

from __future__ import print_function
from . import SigProfilerMatrixGenerator as matGen
import os
import handygenome.sigprofiler_clone.SigProfilerMatrixGenerator as sig
import re
import sys
import pandas as pd
import datetime
from handygenome.sigprofiler_clone.SigProfilerMatrixGenerator.scripts import convert_input_to_simple_files as convertIn
import uuid
import shutil
import time
import numpy as np
import platform
import itertools
import statsmodels
import matplotlib as plt
from pathlib import Path
#import sigProfilerPlotting as sigPlt
import scipy
from handygenome.sigprofiler_clone.SigProfilerMatrixGenerator import install

def perm(n, seq):
	'''
	Generates a list of all available permutations of n-mers.

	Parameters:
			   n  -> length of the desired permutation string
			 seq  -> list of all possible string values

	Returns:
		  permus  -> list of all available permutations
	'''
	permus = []
	for p in itertools.product(seq, repeat=n):
		permus.append("".join(p))
	return(permus)



def SigProfilerMatrixGeneratorFunc (project, genome, vcfFiles, exome=False, bed_file=None, chrom_based=False, plot=False, tsb_stat=False, seqInfo=True, cushion=100, gs=False):
	'''
	Allows for the import of the sigProfilerMatrixGenerator.py function. Returns a dictionary
	with each context serving as the first level of keys. 

	Parameters:
				  project  -> unique name given to the current samples
				   genome  -> reference genome 
				 vcfFiles  -> path where the input vcf files are located.
					exome  -> flag to use only the exome or not
				 bed_file  -> BED file that contains a list of ranges to be used in generating the matrices
			  chrom_based  -> flag to create the matrices on a per chromosome basis
					 plot  -> flag to generate the plots for each context
				 tsb_stat  -> performs a transcriptional strand bias test for the 24, 384, and 6144 contexts. The output is
							  saved into the output/TSB directory
					   gs  -> flag that performs a gene strand bias test

	Returns:
				 matrices  -> dictionary (nested) of the matrices for each context

		example:
			matrices = {'96': {'PD1001a':{'A[A>C]A':23,
										 'A[A>G]A':10,...},
							  'PD1202a':{'A[A>C]A':23,
										 'A[A>G]A':10,...},...},
						'192':{'PD1001a':{'T:A[A>C]A':23,
										 'T:A[A>G]A':10,...},
							  'PD1202a':{'T:A[A>C]A':23,
										 'T:A[A>G]A':10,...},...},...}
	'''
	# verify that installation of all chromosome files was successful
	# 1. get list of all files that were downloaded
	# Terminates the code if the genome reference files have not been created/installed
	lib_loc = os.path.split(os.path.dirname(matGen.__file__))[0]
	tsb_path = os.path.join("references/chromosomes/tsb/", genome)
	if not os.path.exists(os.path.join(lib_loc,tsb_path)):
		print("The specified genome " + genome + " has not been installed\nPlease refer to the SigProfilerMatrixGenerator README for installation instructions:\n\thttps://github.com/AlexandrovLab/SigProfilerMatrixGenerator")
		sys.exit()
	genome_loc = os.path.join("references/chromosomes/transcripts/", genome)
	check_files = [tmp_f for tmp_f in os.listdir(os.path.join(lib_loc, genome_loc)) if not tmp_f.startswith(".")]

	# 2. compare list of downloaded files to md5sum reference
	if genome in list(install.get_reference_md5().keys()):
		for chrom_file_name in check_files:
			chrom_path = os.path.join(lib_loc , tsb_path, chrom_file_name.replace("_transcripts", ""))
			chrom_val = chrom_file_name.strip("_transcripts.txt")
			chrom_md5 = install.md5(chrom_path)
			ref_md5 = install.get_reference_md5()[genome][chrom_val]
			if chrom_md5 != ref_md5:
				print("ERROR: The reference genome installation for", genome, \
					"chromosome", str(chrom_val) + " is incomplete.", \
					"Please run the genome installation script for", genome, \
					"to resolve this issue before proceeding.")
				sys.exit()


	# Instantiates all of the required variables and references
	if not os.path.exists(vcfFiles):
		print("The given project path does not appear to exist. Please check that the specified path exists before proceeding...\n\t" + vcfFiles)
		return()
	if gs:
		print("The Gene Strand Bias is not yet supported! Continuing with the matrix generation.")
		gs = False
	functionFlag = True
	bed = False
	bed_ranges = None
	limited_indel = True
	exome = exome
	plot = plot


	# Instantiates the final output matrix
	matrices = {'96':None, '1536':None, '384':None, '6144':None, 'DINUC':None, '6':None, '24':None, 'INDEL':None}

	# Provides a chromosome conversion from NCBI notation
	ncbi_chrom = {'NC_000067.6':'1', 'NC_000068.7':'2', 'NC_000069.6':'3', 'NC_000070.6':'4', 
				  'NC_000071.6':'5', 'NC_000072.6':'6', 'NC_000073.6':'7', 'NC_000074.6':'8',
				  'NC_000075.6':'9', 'NC_000076.6':'10', 'NC_000077.6':'11', 'NC_000078.6':'12',
				  'NC_000079.6':'13', 'NC_000080.6':'14', 'NC_000081.6':'15', 'NC_000082.6':'16', 
				  'NC_000083.6':'17', 'NC_000084.6':'18', 'NC_000085.6':'19', 'NC_000086.7':'X', 
				  'NC_000087.7':'Y', '82503188|ref|NC_007605.1|':'gi_82503188_ref_NC_007605'}

	# Provides the reference file conversion from binary to base information			  
	tsb_ref = {0:['N','A'], 1:['N','C'], 2:['N','G'], 3:['N','T'],
			   4:['T','A'], 5:['T','C'], 6:['T','G'], 7:['T','T'],
			   8:['U','A'], 9:['U','C'], 10:['U','G'], 11:['U','T'],
			   12:['B','A'], 13:['B','C'], 14:['B','G'], 15:['B','T'],
			   16:['N','N'], 17:['T','N'], 18:['U','N'], 19:['B','N']}

	bias_sort = {'T':0,'U':1,'N':3,'B':2, 'Q':4}
	tsb = ['T','U','N','B']
	tsb_I = ['T','U','N','B','Q']
	bases = ['A','C','G','T']
	mutation_types = ['CC>AA','CC>AG','CC>AT','CC>GA','CC>GG','CC>GT','CC>TA','CC>TG','CC>TT',
					  'CT>AA','CT>AC','CT>AG','CT>GA','CT>GC','CT>GG','CT>TA','CT>TC','CT>TG',
					  'TC>AA','TC>AG','TC>AT','TC>CA','TC>CG','TC>CT','TC>GA','TC>GG','TC>GT',
					  'TT>AA','TT>AC','TT>AG','TT>CA','TT>CC','TT>CG','TT>GA','TT>GC','TT>GG']

	mutation_types_non_tsb = ['AC>CA','AC>CG','AC>CT','AC>GA','AC>GG','AC>GT','AC>TA','AC>TG','AC>TT',
					  'AT>CA','AT>CC','AT>CG','AT>GA','AT>GC','AT>TA',
					  'CG>AT','CG>GC','CG>GT','CG>TA','CG>TC','CG>TT',
					  'GC>AA','GC>AG','GC>AT','GC>CA','GC>CG','GC>TA',
					  'TA>AT','TA>CG','TA>CT','TA>GC','TA>GG','TA>GT',
					  'TG>AA','TG>AC','TG>AT','TG>CA','TG>CC','TG>CT','TG>GA','TG>GC','TG>GT']

	indels_seq_types = [ # Single-sequences
						'C', 'T',

						# Di-sequences
						'AC','AT','CA','CC','CG','CT','GC','TA','TC','TT',

						# Tri-sequences
						'ACC', 'ACT', 'ATC', 'ATT', 'CAC', 'CAT', 'CCA', 'CCC', 'CCG', 'CCT', 'CGC', 'CGT', 'CTA', 'CTC', 'CTG', 'CTT', 
						'GCC', 'GCT', 'GTC', 'GTT', 'TAC', 'TAT', 'TCA', 'TCC', 'TCG', 'TCT', 'TGC', 'TGT', 'TTA', 'TTC', 'TTG', 'TTT',

						# Tetra-sequences
						'AACC', 'AACT', 'AATC', 'AATT', 'ACAC', 'ACAT', 'ACCA', 'ACCC', 'ACCG', 'ACCT', 'ACGC', 'ACGT', 'ACTA', 'ACTC', 'ACTG', 'ACTT', 'AGCC', 'AGCT', 'AGTC', 
						 'AGTT', 'ATAC', 'ATAT', 'ATCA', 'ATCC', 'ATCG', 'ATCT', 'ATGC', 'ATGT', 'ATTA', 'ATTC', 'ATTG', 'ATTT', 'CAAC', 'CAAT', 'CACA', 'CACC', 'CACG', 'CACT', 
						 'CAGC', 'CAGT', 'CATA', 'CATC', 'CATG', 'CATT', 'CCAA', 'CCAC', 'CCAG', 'CCAT', 'CCCA', 'CCCC', 'CCCG', 'CCCT', 'CCGA', 'CCGC', 'CCGG', 'CCGT', 'CCTA', 
						 'CCTC', 'CCTG', 'CCTT', 'CGAC', 'CGAT', 'CGCA', 'CGCC', 'CGCG', 'CGCT', 'CGGC', 'CGTA', 'CGTC', 'CGTG', 'CGTT', 'CTAA', 'CTAC', 'CTAG', 'CTAT', 'CTCA', 
						 'CTCC', 'CTCG', 'CTCT', 'CTGA', 'CTGC', 'CTGG', 'CTGT', 'CTTA', 'CTTC', 'CTTG', 'CTTT', 'GACC', 'GATC', 'GCAC', 'GCCA', 'GCCC', 'GCCG', 'GCCT', 'GCGC', 
						 'GCTA', 'GCTC', 'GCTG', 'GCTT', 'GGCC', 'GGTC', 'GTAC', 'GTCA', 'GTCC', 'GTCG', 'GTCT', 'GTGC', 'GTTA', 'GTTC', 'GTTG', 'GTTT', 'TAAC', 'TACA', 'TACC', 
						 'TACG', 'TACT', 'TAGC', 'TATA', 'TATC', 'TATG', 'TATT', 'TCAA', 'TCAC', 'TCAG', 'TCAT', 'TCCA', 'TCCC', 'TCCG', 'TCCT', 'TCGA', 'TCGC', 'TCGG', 'TCGT', 
						 'TCTA', 'TCTC', 'TCTG', 'TCTT', 'TGAC', 'TGCA', 'TGCC', 'TGCG', 'TGCT', 'TGTA', 'TGTC', 'TGTG', 'TGTT', 'TTAA', 'TTAC', 'TTAG', 'TTAT', 'TTCA', 'TTCC', 
						 'TTCG', 'TTCT', 'TTGA', 'TTGC', 'TTGG', 'TTGT', 'TTTA', 'TTTC', 'TTTG', 'TTTT',

						# Penta-sequences
						'AACCC', 'AACCT', 'AACTC', 'AACTT', 'AATCC', 'AATCT', 'AATTC', 'AATTT', 'ACACC', 'ACACT', 'ACATC', 'ACATT', 'ACCAC', 'ACCAT', 'ACCCA', 'ACCCC', 'ACCCG', 
						'ACCCT', 'ACCGC', 'ACCGT', 'ACCTA', 'ACCTC', 'ACCTG', 'ACCTT', 'ACGCC', 'ACGCT', 'ACGTC', 'ACGTT', 'ACTAC', 'ACTAT', 'ACTCA', 'ACTCC', 'ACTCG', 'ACTCT', 
						'ACTGC', 'ACTGT', 'ACTTA', 'ACTTC', 'ACTTG', 'ACTTT', 'AGCCC', 'AGCCT', 'AGCTC', 'AGCTT', 'AGTCC', 'AGTCT', 'AGTTC', 'AGTTT', 'ATACC', 'ATACT', 'ATATC', 
						'ATATT', 'ATCAC', 'ATCAT', 'ATCCA', 'ATCCC', 'ATCCG', 'ATCCT', 'ATCGC', 'ATCGT', 'ATCTA', 'ATCTC', 'ATCTG', 'ATCTT', 'ATGCC', 'ATGCT', 'ATGTC', 'ATGTT', 
						'ATTAC', 'ATTAT', 'ATTCA', 'ATTCC', 'ATTCG', 'ATTCT', 'ATTGC', 'ATTGT', 'ATTTA', 'ATTTC', 'ATTTG', 'ATTTT', 'CAACC', 'CAACT', 'CAATC', 'CAATT', 'CACAC', 
						'CACAT', 'CACCA', 'CACCC', 'CACCG', 'CACCT', 'CACGC', 'CACGT', 'CACTA', 'CACTC', 'CACTG', 'CACTT', 'CAGCC', 'CAGCT', 'CAGTC', 'CAGTT', 'CATAC', 'CATAT', 
						'CATCA', 'CATCC', 'CATCG', 'CATCT', 'CATGC', 'CATGT', 'CATTA', 'CATTC', 'CATTG', 'CATTT', 'CCAAC', 'CCAAT', 'CCACA', 'CCACC', 'CCACG', 'CCACT', 'CCAGC', 
						'CCAGT', 'CCATA', 'CCATC', 'CCATG', 'CCATT', 'CCCAA', 'CCCAC', 'CCCAG', 'CCCAT', 'CCCCA', 'CCCCC', 'CCCCG', 'CCCCT', 'CCCGA', 'CCCGC', 'CCCGG', 'CCCGT', 
						'CCCTA', 'CCCTC', 'CCCTG', 'CCCTT', 'CCGAC', 'CCGAT', 'CCGCA', 'CCGCC', 'CCGCG', 'CCGCT', 'CCGGC', 'CCGGT', 'CCGTA', 'CCGTC', 'CCGTG', 'CCGTT', 'CCTAA', 
						'CCTAC', 'CCTAG', 'CCTAT', 'CCTCA', 'CCTCC', 'CCTCG', 'CCTCT', 'CCTGA', 'CCTGC', 'CCTGG', 'CCTGT', 'CCTTA', 'CCTTC', 'CCTTG', 'CCTTT', 'CGACC', 'CGACT', 
						'CGATC', 'CGATT', 'CGCAC', 'CGCAT', 'CGCCA', 'CGCCC', 'CGCCG', 'CGCCT', 'CGCGC', 'CGCGT', 'CGCTA', 'CGCTC', 'CGCTG', 'CGCTT', 'CGGCC', 'CGGCT', 'CGGTC', 
						'CGGTT', 'CGTAC', 'CGTAT', 'CGTCA', 'CGTCC', 'CGTCG', 'CGTCT', 'CGTGC', 'CGTGT', 'CGTTA', 'CGTTC', 'CGTTG', 'CGTTT', 'CTAAC', 'CTAAT', 'CTACA', 'CTACC', 
						'CTACG', 'CTACT', 'CTAGC', 'CTAGT', 'CTATA', 'CTATC', 'CTATG', 'CTATT', 'CTCAA', 'CTCAC', 'CTCAG', 'CTCAT', 'CTCCA', 'CTCCC', 'CTCCG', 'CTCCT', 'CTCGA', 
						'CTCGC', 'CTCGG', 'CTCGT', 'CTCTA', 'CTCTC', 'CTCTG', 'CTCTT', 'CTGAC', 'CTGAT', 'CTGCA', 'CTGCC', 'CTGCG', 'CTGCT', 'CTGGC', 'CTGGT', 'CTGTA', 'CTGTC', 
						'CTGTG', 'CTGTT', 'CTTAA', 'CTTAC', 'CTTAG', 'CTTAT', 'CTTCA', 'CTTCC', 'CTTCG', 'CTTCT', 'CTTGA', 'CTTGC', 'CTTGG', 'CTTGT', 'CTTTA', 'CTTTC', 'CTTTG', 
						'CTTTT', 'GACCC', 'GACCT', 'GACTC', 'GACTT', 'GATCC', 'GATCT', 'GATTC', 'GATTT', 'GCACC', 'GCACT', 'GCATC', 'GCATT', 'GCCAC', 'GCCAT', 'GCCCA', 'GCCCC', 
						'GCCCG', 'GCCCT', 'GCCGC', 'GCCGT', 'GCCTA', 'GCCTC', 'GCCTG', 'GCCTT', 'GCGCC', 'GCGCT', 'GCGTC', 'GCGTT', 'GCTAC', 'GCTAT', 'GCTCA', 'GCTCC', 'GCTCG', 
						'GCTCT', 'GCTGC', 'GCTGT', 'GCTTA', 'GCTTC', 'GCTTG', 'GCTTT', 'GGCCC', 'GGCCT', 'GGCTC', 'GGCTT', 'GGTCC', 'GGTCT', 'GGTTC', 'GGTTT', 'GTACC', 'GTACT', 
						'GTATC', 'GTATT', 'GTCAC', 'GTCAT', 'GTCCA', 'GTCCC', 'GTCCG', 'GTCCT', 'GTCGC', 'GTCGT', 'GTCTA', 'GTCTC', 'GTCTG', 'GTCTT', 'GTGCC', 'GTGCT', 'GTGTC', 
						'GTGTT', 'GTTAC', 'GTTAT', 'GTTCA', 'GTTCC', 'GTTCG', 'GTTCT', 'GTTGC', 'GTTGT', 'GTTTA', 'GTTTC', 'GTTTG', 'GTTTT', 'TAACC', 'TAACT', 'TAATC', 'TAATT', 
						'TACAC', 'TACAT', 'TACCA', 'TACCC', 'TACCG', 'TACCT', 'TACGC', 'TACGT', 'TACTA', 'TACTC', 'TACTG', 'TACTT', 'TAGCC', 'TAGCT', 'TAGTC', 'TAGTT', 'TATAC', 
						'TATAT', 'TATCA', 'TATCC', 'TATCG', 'TATCT', 'TATGC', 'TATGT', 'TATTA', 'TATTC', 'TATTG', 'TATTT', 'TCAAC', 'TCAAT', 'TCACA', 'TCACC', 'TCACG', 'TCACT', 
						'TCAGC', 'TCAGT', 'TCATA', 'TCATC', 'TCATG', 'TCATT', 'TCCAA', 'TCCAC', 'TCCAG', 'TCCAT', 'TCCCA', 'TCCCC', 'TCCCG', 'TCCCT', 'TCCGA', 'TCCGC', 'TCCGG', 
						'TCCGT', 'TCCTA', 'TCCTC', 'TCCTG', 'TCCTT', 'TCGAC', 'TCGAT', 'TCGCA', 'TCGCC', 'TCGCG', 'TCGCT', 'TCGGC', 'TCGGT', 'TCGTA', 'TCGTC', 'TCGTG', 'TCGTT', 
						'TCTAA', 'TCTAC', 'TCTAG', 'TCTAT', 'TCTCA', 'TCTCC', 'TCTCG', 'TCTCT', 'TCTGA', 'TCTGC', 'TCTGG', 'TCTGT', 'TCTTA', 'TCTTC', 'TCTTG', 'TCTTT', 'TGACC', 
						'TGACT', 'TGATC', 'TGATT', 'TGCAC', 'TGCAT', 'TGCCA', 'TGCCC', 'TGCCG', 'TGCCT', 'TGCGC', 'TGCGT', 'TGCTA', 'TGCTC', 'TGCTG', 'TGCTT', 'TGGCC', 'TGGCT', 
						'TGGTC', 'TGGTT', 'TGTAC', 'TGTAT', 'TGTCA', 'TGTCC', 'TGTCG', 'TGTCT', 'TGTGC', 'TGTGT', 'TGTTA', 'TGTTC', 'TGTTG', 'TGTTT', 'TTAAC', 'TTAAT', 'TTACA', 
						'TTACC', 'TTACG', 'TTACT', 'TTAGC', 'TTAGT', 'TTATA', 'TTATC', 'TTATG', 'TTATT', 'TTCAA', 'TTCAC', 'TTCAG', 'TTCAT', 'TTCCA', 'TTCCC', 'TTCCG', 'TTCCT', 
						'TTCGA', 'TTCGC', 'TTCGG', 'TTCGT', 'TTCTA', 'TTCTC', 'TTCTG', 'TTCTT', 'TTGAC', 'TTGAT', 'TTGCA', 'TTGCC', 'TTGCG', 'TTGCT', 'TTGGC', 'TTGGT', 'TTGTA', 
						'TTGTC', 'TTGTG', 'TTGTT', 'TTTAA', 'TTTAC', 'TTTAG', 'TTTAT', 'TTTCA', 'TTTCC', 'TTTCG', 'TTTCT', 'TTTGA', 'TTTGC', 'TTTGG', 'TTTGT', 'TTTTA', 'TTTTC', 
						'TTTTG', 'TTTTT']

	# Pre-fills the mutation types variable
	size = 5
	mut_types_initial = perm(size, "ACGT")
	mut_types = []
	for tsbs in tsb:
		for mut in mut_types_initial:
			current_base = mut[int(size/2)]
			if current_base == 'C' or current_base == 'T':
				for base in bases:
					if base != current_base:
						mut_types.append(tsbs+":"+mut[0:int(size/2)] + "[" + current_base+">"+ base+"]"+mut[int(size/2)+1:])


	# Organizes all of the mutation types for DINUCs
	mutation_types_tsb_context = []
	for base in bases:
		for mut in mutation_types:
			for base2 in bases:
				for base3 in tsb:
					mutation_types_tsb_context.append(''.join([base3,":",base,"[",mut,"]",base2]))

	for base in bases:
		for mut in mutation_types_non_tsb:
			for base2 in bases:
				mutation_types_tsb_context.append(''.join(['Q:', base, "[", mut, "]", base2]))


	indel_types_tsb = []
	indel_types_simple = []
	indel_complete = []

	indel_cat = ['Del', 'Ins']

	indel_types = ['1:Del:C:0', '1:Del:C:1', '1:Del:C:2', '1:Del:C:3', '1:Del:C:4', '1:Del:C:5',
				   '1:Del:T:0', '1:Del:T:1', '1:Del:T:2', '1:Del:T:3', '1:Del:T:4', '1:Del:T:5',
				   '1:Ins:C:0', '1:Ins:C:1', '1:Ins:C:2', '1:Ins:C:3', '1:Ins:C:4', '1:Ins:C:5',
				   '1:Ins:T:0', '1:Ins:T:1', '1:Ins:T:2', '1:Ins:T:3', '1:Ins:T:4', '1:Ins:T:5', 
						# >1bp INDELS
				   '2:Del:R:0', '2:Del:R:1', '2:Del:R:2', '2:Del:R:3', '2:Del:R:4', '2:Del:R:5',
				   '3:Del:R:0', '3:Del:R:1', '3:Del:R:2', '3:Del:R:3', '3:Del:R:4', '3:Del:R:5',
				   '4:Del:R:0', '4:Del:R:1', '4:Del:R:2', '4:Del:R:3', '4:Del:R:4', '4:Del:R:5',
				   '5:Del:R:0', '5:Del:R:1', '5:Del:R:2', '5:Del:R:3', '5:Del:R:4', '5:Del:R:5',
				   '2:Ins:R:0', '2:Ins:R:1', '2:Ins:R:2', '2:Ins:R:3', '2:Ins:R:4', '2:Ins:R:5', 
				   '3:Ins:R:0', '3:Ins:R:1', '3:Ins:R:2', '3:Ins:R:3', '3:Ins:R:4', '3:Ins:R:5', 
				   '4:Ins:R:0', '4:Ins:R:1', '4:Ins:R:2', '4:Ins:R:3', '4:Ins:R:4', '4:Ins:R:5',
				   '5:Ins:R:0', '5:Ins:R:1', '5:Ins:R:2', '5:Ins:R:3', '5:Ins:R:4', '5:Ins:R:5',
						#MicroHomology INDELS
				   '2:Del:M:1', '3:Del:M:1', '3:Del:M:2', '4:Del:M:1', '4:Del:M:2', '4:Del:M:3',
				   '5:Del:M:1', '5:Del:M:2', '5:Del:M:3', '5:Del:M:4', '5:Del:M:5', '2:Ins:M:1', 
				   '3:Ins:M:1', '3:Ins:M:2', '4:Ins:M:1', '4:Ins:M:2', '4:Ins:M:3', '5:Ins:M:1', 
				   '5:Ins:M:2', '5:Ins:M:3', '5:Ins:M:4', '5:Ins:M:5', 'complex', 'non_matching']

	for indels in indel_types[:-13]:
		for tsbs in tsb_I:
			indel_types_tsb.append(tsbs + ":" + indels)

	for indels in indels_seq_types:
		repeat = str(len(indels))
		for id_cat in indel_cat:
			for l in range(0, 6, 1):
				indel_complete.append(":".join([repeat, id_cat, indels, str(l)]))
	for id_cat in indel_cat:
		for i in range(0, 6, 1):
			indel_complete.append(":".join(['5',id_cat, '5',str(i)]))

	indel_types_simple = indel_types[:24]
	indel_types_simple.append('long_Del')
	indel_types_simple.append('long_Ins')
	indel_types_simple.append('MH')
	indel_types_simple.append('complex')
	# Instantiates the initial contexts to generate matrices for
	contexts = ['6144']

	# Organizes all of the reference directories for later reference:
	ref_dir, tail = os.path.split(os.path.dirname(os.path.abspath(__file__)))
	chrom_path =ref_dir + '/references/chromosomes/tsb/' + genome + "/"
	if 'havana' in genome:
		genome = genome.split("_")[0]
	transcript_path = ref_dir + '/references/chromosomes/transcripts/' + genome + "/"


	# Terminates the code if the genome reference files have not been created/installed
	if not os.path.exists(chrom_path):
		print("The specified genome " + genome + " has not been installed\nPlease refer to the SigProfilerMatrixGenerator README for installation instructions:\n\thttps://github.com/AlexandrovLab/SigProfilerMatrixGenerator")
		sys.exit()

	# Organizes all of the input and output directories:
	if vcfFiles[-1] != "/":
		vcfFiles += "/"
	vcf_path = vcfFiles + "input/"


	vcf_path_original = vcf_path
	if not os.path.exists(vcf_path) or len(os.listdir(vcf_path)) < 1:
		os.makedirs(vcf_path, exist_ok=True)
		input_files = os.listdir(vcfFiles)
		if os.path.exists(vcfFiles + "input/"):
			input_files.remove("input")
		if os.path.exists(vcfFiles + "logs/"):
			input_files.remove("logs")
		if ".DS_Store" in input_files:
			input_files.remove(".DS_Store")
		if "__init__.py" in input_files:
			input_files.remove("__init__.py")
		if "__pycache__" in input_files:
			input_files.remove("__pycache__")
		if os.path.exists(vcfFiles + "output/"):
			input_files.remove("output")
		for files in input_files:
			shutil.copy(vcfFiles + files, vcf_path + files)
	output_matrix = vcfFiles + "output/"

	if not os.path.exists(output_matrix):
		os.makedirs(output_matrix)


	# Organizes the error and log files
	time_stamp = datetime.date.today()
	output_log_path = vcfFiles + "logs/"
	if not os.path.exists(output_log_path):
		os.makedirs(output_log_path)
	error_file = output_log_path + 'SigProfilerMatrixGenerator_' + project + "_" + genome + str(time_stamp) + ".err"
	log_file = output_log_path + 'SigProfilerMatrixGenerator_' + project + "_" + genome + str(time_stamp) + ".out"

	if os.path.exists(error_file):
		os.remove(error_file)
	if os.path.exists(log_file):
		 os.remove(log_file)
	tempErr = sys.stderr
	sys.stderr = open(error_file, 'w')
	log_out = open(log_file, 'w')
	log_out.write("THIS FILE CONTAINS THE METADATA ABOUT SYSTEM AND RUNTIME\n\n\n")
	log_out.write("-------System Info-------\n")
	log_out.write("Operating System Name: "+ platform.uname()[0]+"\n"+"Nodename: "+ platform.uname()[1]+"\n"+"Release: "+ platform.uname()[2]+"\n"+"Version: "+ platform.uname()[3]+"\n")
	log_out.write("\n-------Python and Package Versions------- \n")
	log_out.write("Python Version: "+str(platform.sys.version_info.major)+"."+str(platform.sys.version_info.minor)+"."+str(platform.sys.version_info.micro)+"\n")
	log_out.write("SigProfilerMatrixGenerator Version: "+sig.__version__+"\n")
	log_out.write("SigProfilerPlotting version: "+sigPlt.__version__+"\n")
	log_out.write("matplotlib version: "+plt.__version__+"\n")
	log_out.write("statsmodels version: "+statsmodels.__version__+"\n")
	log_out.write("scipy version: "+scipy.__version__+"\n")
	log_out.write("pandas version: "+pd.__version__+"\n")
	log_out.write("numpy version: "+np.__version__+"\n")
	

	log_out.write("\n-------Vital Parameters Used for the execution -------\n")
	log_out.write("Project: {}\nGenome: {}\nInput File Path: {}\nexome: {}\nbed_file: {}\nchrom_based: {}\nplot: {}\ntsb_stat: {}\nseqInfo: {}\n".format(project, genome, vcfFiles, str(exome), str(bed_file), str(chrom_based),  str(plot), str(tsb_stat), str(seqInfo)))
	log_out.write("\n-------Date and Time Data------- \n")
	tic = datetime.datetime.now()
	log_out.write("Date and Clock time when the execution started: "+str(tic)+"\n\n\n")
	log_out.write("-------Runtime Checkpoints------- \n")
	log_out.close()



	# Gathers all of the vcf files:
	vcf_files_temp = os.listdir(vcf_path)
	vcf_files = []
	first_extenstion = True
	for file in vcf_files_temp:
		# Skips hidden files
		if file[0:3] == '.DS' or file[0:2] == '__':
			pass
		else:
			vcf_files.append(file)

	# Creates a temporary folder for sorting and generating the matrices
	file_name = vcf_files[0].split(".")
	file_extension = file_name[-1]
	unique_folder = project + "_"+ str(uuid.uuid4())
	output_path = output_matrix + "temp/" + unique_folder + "/"
	if os.path.exists(output_path):
		shutil.rmtree(output_path)
	os.makedirs(output_path)

	skipped_muts = 0
	# Converts the input files to standard text in the temporary folder
	if file_extension == 'genome':
			snv, indel, skipped, samples = convertIn.convertTxt(project, vcf_path, genome, output_path)
	else:
		if file_extension == 'txt':
			snv, indel, skipped, samples = convertIn.convertTxt(project, vcf_path, genome,  output_path, ncbi_chrom, log_file)
		elif file_extension == 'vcf':
			snv, indel, skipped, samples = convertIn.convertVCF(project, vcf_path, genome, output_path, ncbi_chrom, log_file)
		elif file_extension == 'maf':
			snv, indel, skipped, samples = convertIn.convertMAF(project, vcf_path, genome,output_path, ncbi_chrom, log_file)
		elif file_extension == 'tsv':
			snv, indel, skipped, samples = convertIn.convertICGC(project, vcf_path, genome, output_path, ncbi_chrom, log_file)
		else:
			print("File format not supported")

	samples = sorted(samples)
	skipped_muts += skipped
	
	# Instantiates variables for final output statistics
	analyzed_muts = [0, 0, 0]
	
	sample_count_high = 0


	# Begins matrix generation for all possible contexts
	for i in range(0, 2, 1):
		if i == 0 and snv:
			mutation_pd = {}
			mutation_pd['6144'] = pd.DataFrame(0, index=mut_types, columns=samples)
			mutation_dinuc_pd_all = pd.DataFrame(0, index=mutation_types_tsb_context, columns=samples)

			output_path_snv = output_path + "SNV/"
			vcf_files = os.listdir(output_path_snv)
			vcf_path = output_path_snv
			print("Starting matrix generation for SNVs and DINUCs...", end='', flush=True)
			start = time.time()

		# Skips SNVs if none are present
		elif i == 0 and not snv:
			continue
		elif i == 1 and indel:
			mutation_ID = {}
			mutation_ID['ID'] = pd.DataFrame(0, index=indel_types, columns=samples)
			mutation_ID['simple'] = pd.DataFrame(0, index=indel_types_simple, columns=samples)
			mutation_ID['tsb'] = pd.DataFrame(0, index=indel_types_tsb, columns=samples)
			mutation_ID['complete'] = pd.DataFrame(0, index=indel_complete, columns=samples)

			contexts = ['INDEL']
			output_path_indel = output_path + "INDEL/"
			vcf_files = os.listdir(output_path_indel)
			vcf_path = output_path_indel
			print("Starting matrix generation for INDELs...", end='', flush=True)
			start = time.time()

		# Skips INDELs if none are present and deletes the temp folder
		elif i ==1 and not indel:
			shutil.rmtree(output_matrix + "temp/")
			continue

		# Removes hidden files generated in macos
		if ".DS_Store" in vcf_files:
			vcf_files.remove(".DS_Store")


		# Generates the bed regions if a bed file was provided
		if bed_file != None:
			bed = True
			bed_file_path = bed_file
			bed_ranges = matGen.BED_filtering(bed_file_path)
		else:
			bed_file_path = None

		# Sorts files based on chromosome, sample, and start position
		if not chrom_based:
			chrom_start = None
		if i != 1:
			for file in vcf_files:
				chrom = file.split("_")[0]
				if not os.path.exists(chrom_path + chrom + ".txt"):
					continue
				if genome == 'ebv':
					chrom = "_".join([x for x in file.split("_")[:-1]])
				with open(vcf_path + file) as f:
					lines = [line.strip().split() for line in f]
				lines = sorted(lines, key = lambda x: (x[0], int(x[2])))


				context = '6144'
				mutation_pd, skipped_mut, total, total_DINUC, mutation_dinuc_pd_all = matGen.catalogue_generator_single (lines, chrom, mutation_pd, mutation_dinuc_pd_all, mutation_types_tsb_context, vcf_path, vcf_path_original, vcf_files, bed_file_path, chrom_path, project, output_matrix, context, exome, genome, ncbi_chrom, functionFlag, bed, bed_ranges, chrom_based, plot, tsb_ref, transcript_path, tsb_stat, seqInfo, gs, log_file)
				
				if chrom_based and not exome and not bed:
					matrices = matGen.matrix_generator (context, output_matrix, project, samples, bias_sort, mutation_pd, exome, mut_types, bed, chrom, functionFlag, plot, tsb_stat)
					mutation_pd = {}
					mutation_pd['6144'] = pd.DataFrame(0, index=mut_types, columns=samples)
					dinuc_mat = matGen.matrix_generator_DINUC (output_matrix, samples, bias_sort, mutation_dinuc_pd_all, mutation_types_tsb_context, project, exome, bed, chrom, plot)
					mutation_dinuc_pd_all = pd.DataFrame(0, index=mutation_types_tsb_context, columns=samples)


				skipped_muts += skipped_mut
				analyzed_muts[0] += total
				analyzed_muts[1] += total_DINUC

			sample_count_high = len(samples)	

			if exome:
				with open(vcf_path + "exome_temp.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "exome_temp.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()
				mutation_pd = {}
				mutation_pd['6144'] = pd.DataFrame(0, index=mut_types, columns=samples)
				# mutation_pd['6144'], samples2 = matGen.exome_check(mutation_pd['6144'], genome, vcf_path + "exome_temp.txt", output_matrix, project, "SNV", cushion)
				mutation_pd['6144'], samples2 = matGen.exome_check(chrom_based, samples, bias_sort, exome, mut_types, bed, chrom, functionFlag, plot, tsb_stat, mutation_pd['6144'], genome, vcf_path + "exome_temp.txt", output_matrix, project, "SNV", cushion)


			if bed:
				with open(vcf_path + "bed_temp.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "bed_temp.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()


				mutation_pd = {}
				mutation_pd['6144'] = pd.DataFrame(0, index=mut_types, columns=samples)
				mutation_pd['6144'], samples2 = matGen.panel_check(chrom_based, samples, bias_sort, exome, mut_types, bed, chrom, functionFlag, plot, tsb_stat, mutation_pd['6144'], genome, vcf_path + "bed_temp.txt", output_matrix, bed_file_path, project, "SNV", cushion)
				


			if not chrom_based:
				if not mutation_pd['6144'].empty:
					matrices = matGen.matrix_generator (context, output_matrix, project, samples, bias_sort, mutation_pd, exome, mut_types, bed, chrom_start, functionFlag, plot, tsb_stat)
			
			if analyzed_muts[1] > 0:
				if exome:
					with open(vcf_path + "exome_temp_context_tsb_DINUC.txt") as f:
						lines = [line.strip().split() for line in f]
					output = open(vcf_path + "exome_temp_context_tsb_DINUC.txt", 'w')
					for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
							print('\t'.join(line), file=output)
					output.close()

					mutation_dinuc_pd_all = pd.DataFrame(0, index=mutation_types_tsb_context, columns=samples)
					mutation_dinuc_pd_all, samples2 = matGen.exome_check(chrom_based, samples, bias_sort, exome, mutation_types_tsb_context, bed, chrom, functionFlag, plot, tsb_stat, mutation_dinuc_pd_all, genome, vcf_path + "exome_temp_context_tsb_DINUC.txt", output_matrix, project, "DBS", cushion)
			
				if bed:
					with open(vcf_path + "bed_temp_context_tsb_DINUC.txt") as f:
						lines = [line.strip().split() for line in f]
					output = open(vcf_path + "bed_temp_context_tsb_DINUC.txt", 'w')
					for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
						print('\t'.join(line), file=output)
					output.close()

					mutation_dinuc_pd_all = pd.DataFrame(0, index=mutation_types_tsb_context, columns=samples)
					mutation_dinuc_pd_all, samples2 = matGen.panel_check(chrom_based, samples, bias_sort, exome, mutation_types_tsb_context, bed, chrom, functionFlag, plot, tsb_stat, mutation_dinuc_pd_all, genome, vcf_path + "bed_temp_context_tsb_DINUC.txt", output_matrix, bed_file_path, project, "DBS", cushion)

				if not chrom_based:
					if not mutation_dinuc_pd_all.empty:
						dinuc_mat = matGen.matrix_generator_DINUC (output_matrix, samples, bias_sort, mutation_dinuc_pd_all, mutation_types_tsb_context, project, exome, bed, chrom_start, plot)
						matrices['DINUC'] = dinuc_mat

		else:
			for file in vcf_files:
				chrom = file.split("_")[0]
				if genome == 'ebv':
					chrom = "_".join([x for x in file.split("_")[:-1]])
				with open(vcf_path + file) as f:
					lines = [line.strip().split() for line in f]
				lines = sorted(lines, key = lambda x: (x[0], int(x[2])))			
				mutation_ID, skipped_mut, total = matGen.catalogue_generator_INDEL_single (mutation_ID, lines, chrom, vcf_path, vcf_path_original, vcf_files, bed_file_path, chrom_path, project, output_matrix, exome, genome, ncbi_chrom, limited_indel, functionFlag, bed, bed_ranges, chrom_based, plot, tsb_ref, transcript_path, seqInfo, gs, log_file)

				if chrom_based and not exome and not bed:
					matGen.matrix_generator_INDEL(output_matrix, samples, indel_types, indel_types_tsb, indel_types_simple, mutation_ID['ID'], mutation_ID['tsb'], mutation_ID['simple'], mutation_ID['complete'], project, exome, limited_indel, bed, chrom, plot)
					mutation_ID['ID'] = pd.DataFrame(0, index=indel_types, columns=samples)
					mutation_ID['simple'] = pd.DataFrame(0, index=indel_types_simple, columns=samples)
					mutation_ID['tsb'] = pd.DataFrame(0, index=indel_types_tsb, columns=samples)
					mutation_ID['complete'] = pd.DataFrame(0, index=indel_complete, columns=samples)


				sample_count_high = len(samples)
				skipped_muts += skipped_mut
				analyzed_muts[2] += total

			# Performs the final filter on the variants base upon the exome if desired
			if exome:
				with open(vcf_path + "exome_temp.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "exome_temp.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()

				mutation_ID = {}
				mutation_ID['ID'] = pd.DataFrame(0, index=indel_types, columns=samples)
				mutation_ID['ID'], samples2 = matGen.exome_check(chrom_based, samples, bias_sort, exome, indel_types, bed, chrom, functionFlag, plot, tsb_stat, mutation_ID['ID'], genome, vcf_path + "exome_temp.txt", output_matrix, project, "ID", cushion, '83')


				with open(vcf_path + "exome_temp_simple.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "exome_temp_simple.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()

				mutation_ID['simple'] = pd.DataFrame(0, index=indel_types_simple, columns=samples)
				mutation_ID['simple'], samples2 = matGen.exome_check(chrom_based, samples, bias_sort, exome, indel_types_simple, bed, chrom, functionFlag, plot, tsb_stat, mutation_ID['simple'], genome, vcf_path + "exome_temp_simple.txt", output_matrix, project, "ID", cushion, 'simple')


				with open(vcf_path + "exome_temp_tsb.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "exome_temp_tsb.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()

				mutation_ID['tsb'] = pd.DataFrame(0, index=indel_types_tsb, columns=samples)
				mutation_ID['tsb'], samples2 = matGen.exome_check(chrom_based, samples, bias_sort, exome, indel_types_tsb, bed, chrom, functionFlag, plot, tsb_stat, mutation_ID['tsb'], genome, vcf_path + "exome_temp_tsb.txt", output_matrix, project, "ID", cushion, 'tsb')
				mutation_ID['complete'] = pd.DataFrame(0, index=indel_complete, columns=samples)

			if bed:
				with open(vcf_path + "bed_temp.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "bed_temp.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()

				mutation_ID = {}
				mutation_ID['ID'] = pd.DataFrame(0, index=indel_types, columns=samples)
				mutation_ID['ID'], samples2 = matGen.panel_check(chrom_based, samples, bias_sort, exome, indel_types, bed, chrom, functionFlag, plot, tsb_stat, mutation_ID['ID'], genome, vcf_path + "bed_temp.txt", output_matrix, bed_file_path, project, "ID", cushion, '83')

				
				with open(vcf_path + "bed_temp_simple.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "bed_temp_simple.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()

				mutation_ID['simple'] = pd.DataFrame(0, index=indel_types_simple, columns=samples)
				mutation_ID['simple'], samples2 = matGen.panel_check(chrom_based, samples, bias_sort, exome, indel_types_simple, bed, chrom, functionFlag, plot, tsb_stat, mutation_ID['simple'], genome, vcf_path + "bed_temp_simple.txt", output_matrix, bed_file_path, project, "ID", cushion, 'simple')


				with open(vcf_path + "bed_temp_tsb.txt") as f:
					lines = [line.strip().split() for line in f]
				output = open(vcf_path + "bed_temp_tsb.txt", 'w')
				for line in sorted(lines, key = lambda x: (['gi_82503188_ref_NC_007605', 'I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','X','Y','1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23','24',
															'25','26','27','28','29','30','31','32','33','34','35','36','37','38','39', 'MT', 'M', 'MtDNA', 'chrM'].index(x[1]), int(x[2]))):
					print('\t'.join(line), file=output)
				output.close()



				mutation_ID['tsb'] = pd.DataFrame(0, index=indel_types_tsb, columns=samples)
				mutation_ID['tsb'], samples2 = matGen.panel_check(chrom_based, samples, bias_sort, exome, indel_types_tsb, bed, chrom, functionFlag, plot, tsb_stat, mutation_ID['tsb'], genome, vcf_path + "bed_temp_tsb.txt", output_matrix, bed_file_path, project, "ID", cushion, 'tsb')
				mutation_ID['complete'] = pd.DataFrame(0, index=indel_complete, columns=samples)

			if not chrom_based:
				matGen.matrix_generator_INDEL(output_matrix, samples, indel_types, indel_types_tsb, indel_types_simple, mutation_ID['ID'], mutation_ID['tsb'], mutation_ID['simple'], mutation_ID['complete'], project, exome, limited_indel, bed, chrom_start, plot)
				matrices['ID'] = mutation_ID['ID'].iloc[0:83,:]


		if i == 1:
			shutil.rmtree(output_matrix + "temp/")
		end = time.time() - start
		print("Completed! Elapsed time: " + str(round(end, 2)) + " seconds.")

	sys.stderr.close()
	sys.stderr = tempErr
	# Prints a summary for the given run (total samples, skipped mutations, etc.)
	if not chrom_based:
		print("Matrices generated for " + str(sample_count_high) + " samples with " + str(skipped_muts) + " errors. Total of " + str(analyzed_muts[0]) + " SNVs, " + str(analyzed_muts[1]) + " DINUCs, and " + str(analyzed_muts[2]) + " INDELs were successfully analyzed.")
	return(matrices)

	
